/* MausConnect: Nachrichtenformatskonverter fuer das MausNet
   Copyright (C) 1993 1994 1995 Uwe Ohse

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   rfc.c: Allgemeine Routinen fr den Umgang mit dem RFC-Nachrichtenformat
*/


#include "mausconsys.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <time.h>
#include <errno.h>
#include <ctype.h>
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#include <fcntl.h>
#include <syslog.h>
#include <assert.h>
#include <unistd.h>

#include "mauscon.h"
#include "uostr.h" /* vor entumlauten.h (in mausconlib.h) */
#include "mausconlib.h"
#include "metamsg.h"
#include "rfcaddr.h"
#include "rfc.h"
#include "mimesup.h"
#include "822lib.h"

#ifndef isascii
#define isascii(c)    ((((unsigned char)(c)) & ~0x7f) == 0)
#endif

#define STRING(x) #x

#ifndef HAVE_TIMEZONE_VAR
/* oh, well - we *are* in MET, aren't we? */
#define timezone -3600
#endif

static int easy_parse(metalist_t ** ml,char *s);
static void strip_bangpath(char **);

size_t
rfc_head27bit (BOOLEAN mime, uostr_t *to, const char *from)
{
	if (mime)
		return mime_code_headstr (to, from);
	else
	{
		return entumlauten2uostr_add(from,to);
	}
}

size_t
rfc_make_date (char *buf, size_t buflen, uo_sec70_t ti)
{
	size_t l;
	uo_datetime_t dt;
	static const char *days[]= {
			"Thu","Fri","Sat","Sun","Mon","Tue","Wed"
	};

	uo_sec702dt(&dt,&ti);

	l=snprintf(buf,buflen,"%s, %d %s %04ld %02d:%02d:%02d -0000",
		days[(ti/86400)%7],
		dt.day,uo_monabbrev(dt.mon),dt.year,dt.hour,dt.min,dt.sec);
	return l;
}

static size_t
mime_code_head_sub (uostr_t *out, const unsigned char *in)
{
	/* zur Zeit nur quoted printable */
	const unsigned char *pi;
	int fixlen;
	int nextlen;

	pi = in;

	uostr_xadd_cstr(out,"=?ISO-8859-1?Q?");
	nextlen = fixlen = out->len;

	while (*pi)
	{
		if (*pi == '=' || *pi == '?' || *pi == '_' || *pi & 128 || *pi < 32)
		{
			char buf[4];
			sprintf (buf, "=%02X", *pi);
			uostr_xadd_mem(out,buf,3);
			nextlen+=3;
		}
		else if (*pi == ' ')
		{
			uostr_xadd_char(out,'_');
			nextlen++;
		}
		else
		{
			uostr_xadd_char(out,*pi);
			nextlen++;
		}
		if (nextlen > 72)
		{
			uostr_xadd_cstr(out,"?=" LINEEND " =?ISO-8859-1?Q?");
			nextlen = fixlen;
		}
		pi++;
	}
	uostr_xadd_cstr(out,"?=");
	return (out->len);
}

size_t
mime_code_headline (uostr_t *out, const unsigned char *in)
{
	const unsigned char *p;
	size_t l;

	p = in;
	while (*p && *p != ' ' && *p != ':')
		p++;
	if (!*p)
	{
		uostr_xadd_cstr(out,in);
		return out->len;
	}
	l = (p - in) + 1;
	uostr_xadd_mem(out,in,l);
	if (p[1] == ' ') {
		uostr_xadd_char(out,' ');
		p++;
	}
	return mime_code_head_sub (out, p);
}

size_t
mime_code_headstr (uostr_t *out, const unsigned char *in)
{
	/* zur Zeit nur quoted printable */
	const unsigned char *pi;
	int nonasciis;

	pi = in;
	nonasciis = 0;
	while (*pi) {
		if (!isascii (*pi)) {
			nonasciis=1;
			break;
		}
		pi++;
	}
	if (!nonasciis) {
		uostr_t *res;
		res=uostr_xadd_cstr(out,in);
		return res->len;
	}
	return mime_code_head_sub(out,in);
}

int
rfc_metalist2adrheader (BOOLEAN mime, uostr_t *out, metalist_t * list)
{
	metalist_t *pl;

	pl = list;
	uostr_cut(out,0);

	while (pl)
	{
		assert(pl->text);

		xentumlauten2uostr_add(pl->text,out);

		if (pl->ext)
		{
			uostr_xadd_mem(out," (",2);
			rfc_head27bit(mime, out, pl->ext);
			uostr_xadd_char(out,')');
		}
		if (pl->next)
			uostr_xadd_cstr(out,"," LINEEND "\t");
		pl = pl->next;
	}
	uostr_x0(out);
	return out->len-1;
}

/*
 * Eine Headerzeile in einer Nachricht suchen. Vergleichsweise 
 * ineffizient, deshalb nicht fr den Dauergebrauch empfohlen.
 */
char *rfc_find_header(const char *message, const char *headername)
{
	size_t l;
	l=strlen(headername);
	while (*message!='\0' && *message!='\n' && *message!='\r')
	{
		if (message[l]==':'
			&& (message[l+1]==' ' || message[l+1]=='\t')
			&& !strncasecmp(message,headername,l))
		{
			/* Dann ist es der Headername */
			char *p,*p2;
			/* Auf das naechste Zeichen nach "header: " gehen */
			message+=l+2;
			/* Leerzeichen ueberspringen */
			while (*message==' ' || *message=='\t')
				message++;
#if 1
			p=strchr(message,'\n');
			if (!p)
				return XSTRDUP(message);
			else if (p==message) /* yeah - happens sometimes */
				return XSTRDUP("");
			if (p[-1]=='\r')
				p--;
#else
			p=strpbrk(message,"\r\n");
			if (!p)
				return XSTRDUP(message);
#endif
			p2=XMALLOC(p-message+1);
			memcpy(p2,message,p-message);
			p2[p-message]='\0';
			return p2;
		}
#if 1
		message=strchr(message,'\n');
		if (!message)
			return NULL;
		message+=1;
#else
		message=strpbrk(message,"\r\n");
		if (!message)
			return NULL;
		if (*message=='\r')
			message+=2;
		else
			message+=1;
#endif
		if (*message=='\n' || *message=='\r')
			return NULL;
	}
	return NULL;
}

/*
 * Eine Nachricht aus einer Datei lesen, von der wir wissen, da sie n
 * Byte lang ist oder die ganze Nachricht umfat (-1).
 */
int 
rfc_read_message_bytes(int fd, size_t bytes, uostr_t *buf)
{
	char *from;
	char *to;

	int err;
	if (bytes!=(size_t) -1)
	{
		size_t used=buf->len;
		if (buf->size<bytes+used+1) /* +1 wegen \0 */
			uostr_xneed(buf,bytes+1+used);
		
		if (xread(fd,buf->data+buf->len,bytes)==-1)
		{
			err=errno;
			log(LOG_ERR,_("%s: cannot read: %s"),
				"rfc_read_message_bytes",strerror(errno));
			errno=err;
			return -1;
		}
		/* terminate string */
		buf->len+=bytes;
		buf->data[buf->len]='\0';
	}
	else
	{
		/* stat auf Pipes ist nicht so frchterlich gut ... */
#if INT_MAX > 65536
#	define BLOCK 65536U
#else
#	define BLOCK 16384U
#endif
		int count;
		
		do
		{
			if ((size_t)(BLOCK+buf->len+1) > buf->size)
				uostr_xneed(buf,buf->size+BLOCK+1);
			count=read(fd,buf->data+buf->len,BLOCK);
			if (count>0)
				buf->len+=count;
		} while (count>0);
		buf->data[buf->len]='\0';
#undef BLOCK
	}

	/*
	 * Im Folgenden werden die Fortsetzungszeilen zusammengefasst. Das 
	 * ist notwendig, knnte aber auch in rfc_to_meta geschehen. Dort
	 * wre es wahrscheinlich etwas effizienter - aber auch 
	 * erheblicher komplizierter.
	 */
	from=buf->data;
	to=buf->data;
	while (*from)
	{
		/* Achtung: Das funktioniert mit \n und \r\n, und das ist auch
		 * sehr gut so! NNTP/SMTP verwenden \r\n.
		 */
		if (*from=='\n')
		{
			if (from[1]==' ' || from[1]=='\t')
			{
				/* Fortsetzungszeile, \n und folgende Spaces bergehen */
				from++;
				while (from[1]==' ' || from[1]=='\t')
					from++;
				*to=' ';
			}
			else if (from[1]=='\n')
			{
				/* Ende des Headers, nun kopieren */
				size_t l;
				/* Ja, 2 mal */
				*to++=*from++;
				*to++=*from++;
				/* In der Hoffnung auf _optimierte_ strlen/memmove-Routinen */
				l=strlen(from);
				memmove(to,from,l+1);
				return 0;
			}
			else if (from[1]=='\r' && from[2]=='\n')
			{
				/* Ende des Headers, nun kopieren */
				size_t l;
				/* Ja, 3 mal */
				*to++=*from++;
				*to++=*from++;
				*to++=*from++;
				/* In der Hoffnung auf _optimierte_ strlen/memmove-Routinen */
				l=strlen(from);
				memmove(to,from,l+1);
				return 0;
			}
			else
				*to=*from;
		}
		else if (to!=from)
			*to=*from;
		from++;
		to++;
	}
	return 0;

}
/*
 * Eine Datei mit einer rfc-Nachricht darin lesen.
 * *body zeigt hinterher auf den Beginn des Inhalts.
 */
int 
rfc_read_message_fd(int fd,uostr_t  *buf)
{
	struct stat sb;
	int err;

	if (fstat(fd,&sb)==-1)
	{
		/* Wieso das? */
		err=errno;
		log(LOG_WARNING,_("%s: cannot fstat: %s"),"rfc_read_message",
			strerror(errno));
		errno=err;
		return -1;
	}
	if (S_ISFIFO(sb.st_mode) || sb.st_size == 0)
	{
		if (rfc_read_message_bytes(fd,-1,buf))
			return -1;
	}
	else
	{
		if (rfc_read_message_bytes(fd,sb.st_size,buf))
			return -1;
	}
	return 0;
}

/*
 * Eine Datei mit einer rfc-Nachricht darin lesen.
 */
int 
rfc_read_message(const char *fname, uostr_t *buf)
{
	int fd;
	int err;
	int ret;

	fd=open(fname,O_RDONLY);
	if (fd==-1)
	{
		err=errno;
		log(LOG_WARNING,_("%s: cannot open %s: %s"),"rfc_read_message",
			fname, strerror(errno));
		errno=err;
		return -1;
	}
	ret=rfc_read_message_fd(fd,buf);
	close(fd);
	return ret;
}


	/*
	 * Kaputte References gibt es hufig.
	 * Deshalb und wegen CFWS und Consorten ist diese Routine _kein_ Vergngen.
	 */
#define IDSTATE_VOR 0
#define IDSTATE_IN 1 /* < seen */
#define IDSTATE_COMMENT 2 /* unused, commentlevel is used instead */
#define IDSTATE_NFQ 3 /* no-fold-quote */
BOOLEAN
rfc_ids_to_metalist(metalist_t **ml, const char *s, int *troubles)
{
	metalist_t *last;
	BOOLEAN last_ok=FALSE;
	uostr_t id=UOSTR_INIT;
	int commentlevel=0;
	int state=0;
	char smallbuf[100];
	char *sbp=smallbuf;
	if (troubles)
		*troubles=0;
#define COPY(x) do { *sbp++=(x); if (sbp==smallbuf+sizeof(smallbuf)) {uostr_xadd_mem(&id,smallbuf,sizeof(smallbuf));sbp=smallbuf;} } while(0)

	if (!s)
		return last_ok;

	last=*ml;
	while (last && last->next)
		last=last->next;

	while (*s)
	{
		if (commentlevel) {
			/*
				ccontent= ctext / quoted-pair / comment
				comment = "(" *([FWS] ccontent) [FWS] ")"
				CFWS    = *([FWS] comment) (([FWS] comment) / FWS)
			*/
			if (*s==')') {
				commentlevel--;
				s++;
				continue;
			}
			if (*s=='(') {
				commentlevel++;
				s++;
				continue;
			}
			if (*s=='\\') { /* quoted pair */
				if (!s[1]) {
					break;
				}
				s+=2;
				continue;
			}
			/* part of the comment */
			s++;
			continue;
		}
		if (state==IDSTATE_NFQ) {
			/* no-fold-quote =  DQUOTE *(qtext / quoted-pair) DQUOTE
			 * qtext         =  NO-WS-CTL /     ; Non white space controls
			 *                  %d33 /          ; The rest of the US-ASCII
			 *                  %d35-91 /       ;  characters not including "\"
			 *                  %d93-127        ;  or the quote character
			 * der Unterschied zu einem Quoted-String ist, da keine Whitespaces
			 * vorkommen knnen.
			 */
			/* hier knnte man prfen ob das Zeichen wirklich zu den erlaubtem
			 * gehrt.
			 */
			COPY(*s);
			if (*s=='"') {	
				state=IDSTATE_IN;
				s++;
				continue;
			}
			if (*s=='\\') { /* quoted pair */
				if (!s[1])
					break;
				s++;
				COPY(*s);
			}
			s++;
			continue;
		}
		if (state==IDSTATE_VOR) {
			/* msg-id = [CFWS] "<" id-left "@" id-right ">" [CFWS] */
			/* FWS gleich am Stck berspringen */
			while (*s && (*s==' ' || *s=='\t' || *s=='\r' || *s=='\n'))
				s++;
			if (*s=='(') {
				commentlevel++;
				s++;
				continue;
			}
			if (!*s)
				break;
			if (*s!='<') {
				/* iirgendein Mll ... berspringen */
				s++;
				last_ok=FALSE;
				if (troubles)
					(*troubles)++;
				continue;
			}
			s++; /* don't copy */
			state=IDSTATE_IN;
			sbp=smallbuf;
			uostr_fcut(&id);
			continue;
		}
		/* msg-id = [CFWS] "<" id-left "@" id-right ">" [CFWS] */
		/* id-left= dot-atom-text / no-fold-quote / obs-id-left *
		 * no-fold-quote = DQUOTE *(qtext / quoted-pair) DQUOTE
		 * dot-atom-text = 1*atext *("." 1*atext)
		 * obs-id-left = local-part
		 * local-part  = dot-atom / quoted-string / obs-local-part
		 * quoted-string=[CFWS] DQUOTE *([FWS] qcontent) [FWS] DQUOTE [CFWS]
		 * dot-atom     =[CFWS] dot-atom-text [CFWS]
		 * obs-local-part=atom *("." atom)
		 * CFWS =       *([FWS] comment) (([FWS] comment) / FWS)
		 * FWS  =       ([*WSP CRLF] 1*WSP) /   ; Folding white space
		 *              obs-FWS
		 * obs-FWS =    1*WSP *(CRLF 1*WSP)
		 * Runs of CFWS in a structured field header are semantically interpreted
		 * as a single space character.
		 * 
		 * Hochgradig hlich.
		 * wir machen es uns einfach - innerhalb der ID werden nur 
		 * ".." und \x erkannt, der rest 1:1 kopiert. 
		 */
		/* state==IDSTATE_IN */
		if (*s == '"') {
			COPY(*s);
			s++;
			state=IDSTATE_NFQ;
			continue;
		}
		if (*s=='\\') {
			if (!s[1]) 
				break;
			COPY(*s);
			s++;
			COPY(*s);
			s++;
			continue;
		}
		if (*s=='>') {
			if (sbp>smallbuf) {
				metalist_t *new;
				uostr_xadd_mem(&id,smallbuf,sbp-smallbuf);
				sbp=smallbuf;
				uostr_x0(&id);
				new=XMALLOC(sizeof(metalist_t));
				/* reuse that uostr. We need to free that memory anyway */
				new->text=id.data;
				uostr_forget(&id);
				id.data=0;
				new->ext=NULL;
				new->next=NULL;
				if (last)
				{
					last->next=new;
					last=new;
				}
				else
				{
					*ml=new;
					last=new;
				}
				uostr_fcut(&id);
				s++;
				last_ok=TRUE;
				state=IDSTATE_VOR;
				continue;
			}
			/* Hmm??? <>??? */
			last_ok=FALSE;
			state=IDSTATE_VOR;
			if (troubles)
				(*troubles)++;
			continue;
		}
		if (*s=='<') {
			s++;
			last_ok=FALSE;
			state=IDSTATE_VOR;
			continue;
		}
		COPY(*s);
		s++;
		continue;
	}
	uostr_freedata(&id);
	return last_ok;
}

/* XXX: USEFOR erlaubt unicode fr Gruppennamen - wir nicht.
 */
#define GSTATE_VOR 0
#define GSTATE_IN 1 /* < seen */
#define GSTATE_COMMENT 2 /* unused, commentlevel is used instead */
#define GSTATE_NFQ 3 /* no-fold-quote */
/*
 * Newsgroups-content = newsgroup-name *( ng-delim newsgroup-name)
 * newsgroup-name = *FWS component *( "." component ) *FWS
 * component = component-start [*component-rest component-start]
 * component-start = lowercase / digit
 * lowercase = <Unicode Letter, Lowercase> / <Unicode Letter, Other>
 * uppercase = <Unicode Letter, Uppercase> / <Unicode Letter, Titlecase>
 * digit = <Unicode Number, Decimal Digit> / <Unicode Number, Other>
 * component-rest = component-start / "+" / "-" / "_"
 * ng-delim = ","
 */
static BOOLEAN
rfc_groups_to_metalist(metalist_t **ml,char *s)
{
	metalist_t *last;
	metalist_t *new;
	char *p;
	BOOLEAN ok=TRUE;

	if (!s)
		return FALSE;

	/* append to end of list */
	last=*ml;
	while (last && last->next)
		last=last->next;

	while (TRUE) {
		char *q;
		char c=0;
		/* skip white space */
		while (*s==' ' || *s=='\t' || *s=='\r' || *s=='\n')
			s++;
		if (!*s) {
			ok=FALSE;
			break;
		}
		p=strchr(s,',');
		if (p==s) {
			/* seufz. Mll. Generiere Warnung */
			ok=FALSE;
			s=p+1;
			continue;
		}
		if (p)
			*p=0;
		for (q=s;*q && *q!=' ' && *q!='\t' && *q!='\r' && *q!='\n';q++)
			/* nothing */;
		c=*q;
		if (*q) {
			/* space after (ok) or in (forbidden) newsgroup name */
			char *r;
			*q=0;
			/* skip over the whitespaces */
			for (r=q+1;*r && (*r==' ' || *r=='\t' || *r=='\r' || *r=='\n');r++)
				/* nothing */;
			if (*r)
				ok=FALSE;
		}
		new=XMALLOC(sizeof(metalist_t));
		new->text=XSTRDUP(s);
		*q=c;
		new->ext=NULL; 
		new->next=NULL;
		if (last)
			last->next=new;
		else
			*ml=new;
		last=new;
		if (!p)
			break;
		*p=',';
		s=p+1;
	}
	return ok;
	
}

#define EASY_FIRST
int 
rfc_adress_to_metalist(int meckern,metalist_t **ml,char *s)
{
	metalist_t *last;
	metalist_t *new=NULL;
#if 0
	metalist_t *akt;
#endif
	rp_addr_t *ap;
	int fails;
	char **dom;

	int ep_done=0; /* Easy-Parse tatschlich aufgerufen? */

	last=*ml;
	while (last && last->next)
		last=last->next;

#ifdef EASY_FIRST
	/* Nach Mglichkeit easy_parse benutzen, da diese Funktion wesentlich
	 * schneller ist.
	 */
	{
		char *at=strchr(s,'@');
		char *p;
		/* Wenn ':' oder ',' vor dem at-Zeichen kommen ist die 
		 * Adresse eine Sourceroute nach RFC82[12]. Die kann 
		 * easy_parse nicht.
		 * Kommen : oder , nach dem @, ist die Adresse irgendetwas
		 * anderes - z.B. irgendein Fidozeugs.
		 */
		if (at && ((p=strchr(at+1,'@'))==NULL) &&
			((p=strpbrk(s,",:"))==NULL || p>at))
		{
			if (0==easy_parse(ml,s))
				return 0; /* fine, no trouble */
			ep_done=1;
		}
	}
#endif

	rfc_parse_address(s);

	fails=0;

	ap=adrlist;
	while (ap)
	{
		if (ap->localp)
		{
			uostr_t tmp=UOSTR_INIT;
			new=XMALLOC(sizeof(metalist_t));
			if (ap->name)
				new->ext=XSTRDUP(ap->name);
			else if (ap->comment)
				new->ext=XSTRDUP(ap->comment);
			else
				new->ext=NULL;

			uostr_xdup_cstr(&tmp,ap->localp);

			if (ap->destdom) {
				uostr_xadd_char(&tmp,'@');
				for (dom = ap->destdom->sub; dom != ap->destdom->top; dom++) {
					uostr_xadd_cstr(&tmp,*dom);
					uostr_xadd_char(&tmp,'.');
				}
				uostr_xadd_cstr(&tmp,*dom);
				uostr_x0(&tmp);
			}
			else
			{
				char *p;
				uostr_x0(&tmp);
				/* da ist jetzt mglicherweise keine Domain, oder 
				 * eine in uucp-Schreibweise. Das knnen wir fixen.
				 */
				p=strchr(tmp.data,'%');
				if (p)
					*p='@';
				else if ((p=strchr(ap->localp,'!'))!=NULL)
				{
					*p=0;
					uostr_xdup_cstr(&tmp,p+1);
					uostr_xadd_char(&tmp,'@');
					uostr_xadd_cstr(&tmp,ap->localp);
					uostr_x0(&tmp);
					*p='!';
				}
			}
			new->text=tmp.data;
			uostr_forget(&tmp);

			new->next=NULL;
			if (last)
			{
				last->next=new;
				last=new;
			}
			else
			{
				*ml=new;
				last=new;
			}
		}
		else
			fails++;
		ap=ap->next;
	}
	if (errlist)
		fails++;

	if (fails && !new)
	{
		if (ep_done || easy_parse(ml,s))
		{
			if (meckern)
				log(LOG_WARNING,_("could not parse %s"),s);
			return fails;
		}
	}
	if (fails && meckern) {
		ap=errlist;
		log(LOG_INFO,"trouble parsing %s",s);
		while (ap) {
			if (ap->error)
				log(LOG_NOTICE,"%s",ap->error);
			ap=ap->next;
		}
	}
	return fails;
}

static int 
easy_parse(metalist_t ** ml,char *s)
{
	/* der zweite Versuch, mit einer Adresse klarzukommen */
	char *p;
	metalist_t *new=NULL;
	metalist_t *last;
	p=strpbrk(s,",;[]\"");
	if (p)
		return 1;
	p=strchr(s,'<');
	if (p)
	{
		/* "Irgendein Text" <adresse> 
		 * oder
		 *  Mein Name <adresse> 
		 */
		char *p1; /* Zeiger auf > */
		char *p2; /* Zeiger auf Beginn des Realnames */
		char *p3; /* Zeiger auf Ende des Realnames */
		p1=strchr(p,'>');
		if (!p1)
			return 1;
		new=XMALLOC(sizeof(metalist_t));
		/* 
		 * Vor nach <alpha@beta>
		 *          p          p1
		 */
		new->text=XMALLOC(p1-p);
		memcpy(new->text,p+1,p1-p-1);
		new->text[p1-p-1]='\0';
		p2=s;
		while (isspace(*p2))
			p2++;
		if (p2!=p)
		{
			p3=p-1;
			while (isspace(*p3) && p3>p2)
				p3--;
			/* Vor nach <alpha@beta> 
		     * p2     p3
		     * 0      7  Platzbedarf fr "Vor Nach" ist 9.
		     */
			new->ext=XMALLOC(p3-p2+2);
			memcpy(new->ext,p2,p3-p2+1);
			new->ext[p3-p2+1]='\0';
		}
		else
			new->ext=NULL;
		new->next=NULL;
	}
	else
	{
		/* Dann ist es hoffentlich die andere verbreitete Form,
		 * adresse (name) , mit (name) adresse kommen wir aber auch
		 * klar.
		 */
		char *p1;       /* Zeiger auf ( */
		char *p2;       /* Zeiger auf ) */
		char *dot;

		/* A . (B) C @ D -> A.C@D, aber das geht mit dieser Routine hier nicht ... */
		dot=strpbrk(s,"@.");
		if (dot && *dot=='.')
			return 1;
		p=s;
		while (isspace(*p))
			p++;
		p1=p;
		while (*p1 && *p1!='(')
			p1++;
		if (*p1=='(')
		{
			p2=p1;
			while (*p2 && *p2!=')')
				p2++;
			if (!*p2)
				return 1; /* nichts zu machen */
			if (p1!=p)
			{
				/* adresse (name) */
				char *p4; /* Zeiger auf Ende der Adresse */
				int namel;
				p4=p1-1;
				while (isspace(*p4))
					p4--;
				new=XMALLOC(sizeof(metalist_t));

				/*
				 * Adresse  , Laenge 7, Platzbedarf 8
				 * p1    p4
				 * 0     6
				 */
				new->text=XMALLOC(p4-p+2);
				memcpy(new->text, p, p4-p+1);
				new->text[p4-p+1]='\0';

				/*
				 * (name) , Laenge 6, Laenge name 4, Platzbedarf 5
				 * p1   p2
				 * 0    5
				 */
				namel=p2-p1-1;
				if (namel>0) /* () ist gar nicht so selten */
				{
					new->ext=XMALLOC(namel+1);
					memcpy(new->ext, p1+1, namel);
					new->ext[namel]='\0';
				}
				else 
					new->ext=NULL;

				new->next=NULL;
			}
			else
			{
				/* (name) adresse */
				char *p3; /* Zeiger auf Beginn der Adresse */
				int namel;
				p3=p2+1;
				while (isspace(*p3))
					p3++;
				if (!*p3)
					return 1; /* Keine Adresse angegeben */

				new=XMALLOC(sizeof(metalist_t));

				new->text=XSTRDUP(p3);

				namel=p2-p1-1;
				if (namel>0) /* () ist gar nicht so selten */
				{
					new->ext=XMALLOC(namel+1);
					memcpy(new->ext, p1+1, namel);
					new->ext[namel]='\0';
				}
				else 
					new->ext=NULL;

				new->next=NULL;
			}
		}
		else
		{
			/* Dann kann es nur eine einfache Adresse sein */
			new=XMALLOC(sizeof(metalist_t));
			new->text=XSTRDUP(p);
			p=new->text+strlen(new->text)-1;
			while (isspace(*p))
				*p--='\0';
			new->ext=NULL;
			new->next=NULL;
		}
	}
#if 1
	if (*new->text=='\"')
	{
		char *p_anfz;
		p_anfz=strchr(new->text+1,'\"');
		if (p_anfz)
		{
			if (p_anfz==new->text+1) {
				memmove(new->text,p_anfz+1,strlen(p_anfz+1)+1);
			} else {
				memmove(new->text,new->text+1,p_anfz-new->text-1);
				memmove(p_anfz-1,p_anfz+1,strlen(p_anfz+1)+1);
			}
		}
	}
#endif

	last=*ml;
	while (last && last->next)
		last=last->next;
	if (last)
		last->next=new;
	else
		*ml=new;
	return 0;
}

/* 
 * Grrr - mauskompatibilitt der umfangreicheren Art:
 * kiste!user@pollsite     -> user@kiste
 * k2!k1!user@pollsite     -> user@k1
 * k2!k1!user ->user@k1
 * kiste!user ->user@kiste
 * Das ignoriert Teile der Problematik!
 */
static void
strip_bangpath(char **adr)
{
	char *bang;
	char *at,*p;
	char *res;
	char *domain,*user;
	char *start=*adr;
	bang=strchr (start,'!');
	if (!bang)
		return;
	if (bang==start)
		return;

	at=strchr(start,'@');
	if (at && bang >at)
		return;
	if (at) {
		/*
		 * k2!k1!user@somewhere -> k2!k1!user
		 * kiste!user@somewhere -> kiste!user
		 */
		*at=0;
	}

	p=bang+1;
	domain=start;
	user=bang;
	while (*p) {
		if (*p=='!') {
			if (!user) {
				user=p;
			} else {
				domain=user+1;
				user=p;
			}
		}
		p++;
	}
	*user++=0;

	res=concat(user,"@",domain,NULL);
	free(*adr);
	XNULLCHECK(res);
	*adr=res;
	return;
}

int 
rfc_to_meta(uostr_t *buf,metamsg_t *meta, int is_mail)
{
	char *p;
	char *linep;          /* Beginn der Zeile */
	char *namep;
	size_t namel;
	char *contentp;       /* Zeiger auf Inhalt einer Headerzeile */
	int errors_detected=0;
	int in_header=1;
	
	enum mime_encoding enc=seven_bit;
	BOOLEAN can_decode_mime_encodings=FALSE;
	int mime_version_major=0;
	int mime_version_minor=0;
	char *return_path=0;
	char *delivered_to=0;

	metalist_t *inrepto=NULL;

	/* die Felder in der Nachricht lschen */
	meta_clean_message (meta);
	meta->is_mail=is_mail;

	/* Nachrichtenverarbeitung */
	linep=buf->data;

	/* 
	 * Header verarbeiten: Dies kann auf zwei Arten geschehen. 
	 * 1. man durchluft alle Headerzeilen und sucht den Namen in 
	 *    der Liste der bekannten Namen.
	 * 2. man sucht im Header nach allen interessanten (bekannten / zu
	 *    verarbeitenden) Strings.
	 * Die meisten Nachrichten haben nur wenige Headerzeilen, von daher
	 * ist 1. wohl schneller.
	 *
	 * Gesplittete Headerzeilen sind vorher schon zusammengesetzt worden
	 * (rfc_read_message_bytes).
	 */
	while (linep && *linep!='\0' && in_header)
	{
		namep=linep;
		/* 
		 * Zeile mit \0 statt \n terminieren und linep auf den Beginn der
		 * naechsten Zeile setzen.
		 */
		/* p=strpbrk(namep,"\r\n"); */
		contentp=0;
		for (p=namep;*p && *p!='\r' && *p!='\n'; p++) {
			if (*p==':' && !contentp) 
				contentp=p;
		}
		if (p) {
			/* linep ist von nun an der Zeiger auf die nchste Zeile */
			if (*p=='\r') {
				/* Es gab mal diesen Mist: "From: <....\r>\n", plus coredump */
				if (p[1]=='\n')
					linep=p+2;
				else
				{
					linep=strchr(p,'\n');
					if (linep)
						linep++;
				}
			}
			else
				linep=p+1;
			*p='\0';
		}
		else
			linep=NULL; /* Keine weitere Zeile */
		if (*namep=='\0')
		{
			/* Leerzeile beendet Header */
			size_t l;
			
			in_header=0;

			if (mime_version_major>=1)
			{
				/* wir sind freundlich und decodieren, was gefahrlos zu
				 * decodieren ist: text/irgendwas 
				 */
				if ((enc==base_64 || enc==quoted_printable)
					&& can_decode_mime_encodings)
				{
					mime_decode_buffer(linep,enc);
					/* die Nachricht hat keine Codierung mehr */
					free(meta->content_transfer_encoding);
					meta->content_transfer_encoding=NULL;
				}
			}

			/* Nun aus dem Lesepuffer in die metamsg_t-Struktur kopieren */
			l=strlen(linep);
			if (l+1>meta->body.size)
			{
				size_t lneu;

				lneu=l+64*1024; /* nicht kleckern, das rentiert sich nicht */
				uostr_xneed(&meta->body,lneu);
			}
			meta->body.len+=l;
			memcpy(meta->body.data,linep,l+1);
			meta->body.data[meta->body.len]=0;
			continue;
		}
		if (!contentp)
		{	
			/* Was? Wie bitte? Zeile ohne Doppelpunkt? Im Header? */
			continue;
		}
		namel=contentp-namep-1;
		*contentp++='\0';
		if (*contentp==' ' || *contentp=='\t')
			contentp++;
		if (!*contentp) {
			continue; /* leere Zeile wie "keine Zeile" behandeln */
		}

		/* MIME in Headerzeilen dekodieren */
		p=contentp;
		while (*p && (p[0]!='=' || p[1]!='?'))
			p++;
		if (*p)
			mime_decode_headerline(p);

		/* simple Form von Hashing, aber effizient */
		switch(tolower(*(namep++)))
		{
		case 'a':
			/* Also-Control */
			if (!strhalfcasecmp(namep,"lso-control")) {
				meta->also_control=XSTRDUP(contentp);
				meta->mausflags.exportflag=XR_MINUS;
				meta_add_trace(meta,MMT_CONTROL|MMT_NX,
					"also-control found: %s",contentp);
				continue;
			} else if (!strhalfcasecmp(namep,"rchive-name")) {
				meta_add_trace(meta,MMT_LOSS|MMT_NX,
					"archive-name found: %s",contentp);
				meta->mausflags.exportflag=XR_MINUS;
			} else if (!strhalfcasecmp(namep,"pproved")) {
				meta_add_trace(meta,MMT_LOSS|MMT_NX,
					"approved found: %s",contentp);
				meta->mausflags.exportflag=XR_MINUS;
			} else if (!strhalfcasecmp(namep,"rticle-names")) {
				meta_add_trace(meta,MMT_LOSS|MMT_NX,
					"article-names found: %s",contentp);
				meta->mausflags.exportflag=XR_MINUS;
			} else if (!strhalfcasecmp(namep,"rticle-updates")) {
				meta_add_trace(meta,MMT_LOSS|MMT_NX,
					"article-updates found: %s",contentp);
				meta->mausflags.exportflag=XR_MINUS;
			}
			break;
		case 'b':
			if (!strhalfcasecmp(namep,"cc")) {
				/* Fehler merken, aber dennoch weitermachen */
				if (rfc_adress_to_metalist(1,&meta->bcc,contentp)!=0 || !meta->bcc) {
					meta->mausflags.exportflag=XR_MINUS;
					meta_add_trace(meta,MMT_ERROR|MMT_NX, 
						"problems converting 'bcc' header");
					/* deliberately choose to not put it in an "unknown" header */
				}
				continue;
			}
			break;
		case 'c':
			/* Control */
			if (!strhalfcasecmp(namep,"ontrol")) {
				meta->control=XSTRDUP(contentp);
				meta->mausflags.exportflag=XR_MINUS;
				meta_add_trace(meta,MMT_CONTROL|MMT_NX,
					"control found: %s",contentp);
				continue;
			} else if (!strhalfcasecmp(namep,"c")) {
				/* Fehler merken, aber dennoch weitermachen */
				if (rfc_adress_to_metalist(1,&meta->cc,contentp)!=0 || !meta->cc) {
					meta->mausflags.exportflag=XR_MINUS;
					meta_add_trace(meta,MMT_ERROR|MMT_NX, 
						"problems converting 'cc' header");
					meta_add_list (&meta->unknown_headers,
						"X-Mauscon-Original-Cc",
						contentp, 0);
				}
				continue;
			} else if (!strhalfcasecmp(namep,"ontent-type")) {
				/* text/plain und text -> kein Content-Type. Ist eh' default */
				char *next;
				char *ct=rfc_getword(contentp,&next);
				int more_in_it=FALSE;
				int is_8859_1=FALSE;
				if (!ct || 0==strncasecmp("text/",ct,5) 
					|| 0==strcasecmp("text",ct))
				{
					char *valtag;
					while ((valtag=rfc_getword(next,&next))!=NULL) {
						if (0==strcasecmp(valtag,"charset=us-ascii")
							|| 0==strcasecmp(valtag,"charset=iso-8859-1")) {
							is_8859_1=TRUE;
						} else 
							more_in_it=TRUE;
						free(valtag);
					}
				}
				/* we can do MIME q-p/b64 decoding for any text type */
				if (is_8859_1) can_decode_mime_encodings=1;
				/* but don't want Content-Type lines for Content-Type "text"
				 * and "text/plain" if the lines don't contain more information */
				if (ct && (more_in_it ||
					(0!=strcasecmp(ct,"text") && 0!=strcasecmp(ct,"text/plain"))))
					meta->content_type=XSTRDUP(contentp);
				if (ct)
					free(ct);

#ifdef WANT_CS_WANDLUNG
				/* but not yet, as it doesn't work */
				if (is_text)
				{
				    char *p_look;
					p_look=contentp;
					while (*p_look && strncasecmp(p_look,"charset",7))
						p_look++;
					if (*p_look)
					{
						p_look+=strlen("charset");
						while (*p_look && *p_look!='=')
							p_look++;
						if (*p_look=='=')
						{
							char tmpbuf[128];
							char *q;

							p_look++;

#error bufferoverlow, do not use me
							q=strchr(p_look,';');
							if (!q)
								strcpy(tmpbuf,p_look);
							else
								strncpy(tmpbuf,p_look,q-p_look);
							tmpbuf[q-p_look]='\0';
							strtrim(tmpbuf);
							cs=cslib_get_charset_from_name(tmpbuf);
						}
					}
				}
#endif
				continue;
			} else if (!strhalfcasecmp(namep,"ontent-description")) {
				meta->content_description=XSTRDUP(contentp);
				continue;
			} else if (!strhalfcasecmp(namep,"ontent-id")) {
				meta->content_id=XSTRDUP(contentp);
				continue;
			} else if (!strhalfcasecmp(namep,"ontent-transfer-encoding")) {
				enc=mime_get_encoding(contentp);
				meta->content_transfer_encoding=XSTRDUP(contentp);
				continue;
			} else if (!strnhalfcasecmp(namep,"ontent-",7)) {
				/* unbekannte Mime-Zeilen */
				/* meta->mausflags.noexport=TRUE; */
				meta_add_list (&meta->unknown_headers,namep-1,contentp,0);
			}
			break;
		case 'd':
			/* Distribution */
			/* den Standardfehler, "world" mitzusenden, abfangen */
			if (!strhalfcasecmp(namep,"istribution")
				&& strcmp(contentp,"world"))
			{
				meta->distribution=XSTRDUP(contentp);
				if (opts.workmode!=USER) {
					meta->mausflags.exportflag=XR_MINUS;
					meta_add_trace(meta,MMT_DIST|MMT_NX,"distribution %s found",contentp);
				}
				continue;
			}
			/* Date */
			if (!strhalfcasecmp(namep,"ate")) {
				/* 
				 * getdate funktioniert auch dann, wenn Date leer ist, und 
				 * setzt dann das aktuelle Datum, aber Uhrzeit 0:0 ein.
				 * Bei Verwendung einer anderen Routine muss das Verhalten
				 * geprueft werden.
				 */
				int er;
				uostr_t x;

				x.data=contentp;
				x.len=strlen(contentp);
				er=r822_date(&meta->date,&x);
				if (er==R822_ERR_NOMEM)
					XNULLCHECK(0);
				else if (er==R822_ERR_PARSE) {
					meta->date=-1;
					meta->mausflags.exportflag=XR_MINUS;
					meta_add_trace(meta,MMT_ERROR|MMT_NX,
						"couldn't convert date %s",
						contentp);
				}
				/* Weiter oben wird der Rest abgefangen */
				continue;
			}
			/* delivered-to */
			if (!strhalfcasecmp(namep,"elivered-to") && !delivered_to) {
				metalist_t *ml=NULL;
				if (rfc_adress_to_metalist(1,&ml,contentp)!=0 || !ml) {
					meta_add_trace(meta,MMT_ERROR, 
						"problems converting 'delivered-to'");
					meta_add_list (&meta->unknown_headers,
						"X-Mauscon-Original-Delivered-To",
						contentp, 0);
				} else {
					if (ml && ml->text)
						delivered_to=XSTRDUP(ml->text);
					if (ml->next) {
						meta_add_trace(meta,MMT_ERROR,
							"zu viele envelope-Empfnger: %s", contentp);
						meta_add_list (&meta->unknown_headers,
							"X-Mauscon-Original-Delivered-To",
							contentp, 0);
					}
				}
				meta_free_list(&ml);
				continue;
			}
			break;
		case 'e':
			if (!strhalfcasecmp(namep,"xpires")) {
				meta->mausflags.exportflag=XR_MINUS;
				meta_add_trace(meta,MMT_CONTROL|MMT_NX,
					"expires found: %s", contentp);
			}
			break;
		case 'f':
			/* From: */
			if (!strhalfcasecmp(namep,"rom"))
			{
				/* Fehler merken, aber dennoch weitermachen */
				if (rfc_adress_to_metalist(1,&meta->head_from,contentp)!=0 || !meta->head_from) {
					meta->mausflags.exportflag=XR_MINUS;
					meta_add_trace(meta,MMT_ERROR|MMT_NX, 
						"problems converting 'from' header");
					meta_add_list (&meta->unknown_headers,
						"X-Mauscon-Original-From",
						contentp, 0);
				}
				continue;
			}
			/* Followup-To: */
			else if (!strhalfcasecmp(namep,"ollowup-to"))
			{
				if (!rfc_groups_to_metalist(&meta->followup_to,contentp)) {
					meta->mausflags.exportflag=XR_MINUS;
					meta_add_trace(meta,MMT_ERROR|MMT_NX,
						"Followup-To unwandelbar: %s", contentp);
					meta_add_list (&meta->unknown_headers,
						"X-Mauscon-Original-Followup-To",
						contentp, 0);
				}
				if (!meta->followup_to)
					errors_detected++;
				continue;
			}
			break;
		case 'i':
			/* in-reply-to */
			if (!strhalfcasecmp(namep,"n-reply-to"))
			{
				char *kl;
				char *at;
				char *gr;
				char *sp;
				if (!is_mail)
					continue; /* in-reply-to ist fr News sinnlos */
				kl=contentp-1;
				while ((kl=strchr(kl+1,'<'))!=NULL)
				{
					/* kl zeigt auf < */
					gr=strchr(kl,'>');
					if (!gr)
						continue;
					at=strchr(kl,'@');
					if (!at)
						continue;
					if (at<gr)
					{
						sp=strpbrk(kl," \t");
						if (!sp || sp>gr)
						{
							*gr='\0';
							inrepto=XMALLOC(sizeof(metalist_t));
							inrepto->text=XSTRDUP(kl+1);
							inrepto->next=NULL;
							inrepto->ext=NULL;
							*gr='>';
							continue;
						}
						break;
					}
				}
				continue;
			}
		case 'k':
			/* Keywords */
			if (!strhalfcasecmp(namep,"eywords")) {
				meta->keywords=XSTRDUP(contentp);
				continue;
			}
			break;
		case 'm':
			/* Message-ID: */
			if (!strhalfcasecmp(namep,"essage-id")) {
				metalist_t *ml=NULL;
				int troubles=0;
				if (!rfc_ids_to_metalist(&ml,contentp,&troubles)||troubles) {
					char *lastchar;
					/* die ist defekt ... */
					meta->mausflags.exportflag=XR_MINUS;
					meta_add_trace(meta,MMT_ERROR|MMT_NX,
						"Message-ID unwandelbar: %s", contentp);
					meta_add_list (&meta->unknown_headers,
						"X-Mauscon-Original-Message-Id",
						contentp, 0);
					if (!is_mail)
						errors_detected++;
					/* last try: fr Mail und Logfiles */
					lastchar=contentp+strlen(contentp)-1;
					while (isspace(*lastchar))
						*lastchar--='\0';
					while (isspace(*contentp))
						contentp++;
					if (*lastchar=='>')
						lastchar--;
					if (!*lastchar)
						continue; /* leere message-id? Wird zusammen mit "keine Message-ID" behandelt */
					if (*contentp=='<')
						contentp++;
					if (contentp>lastchar) /* <> */
						continue;
					meta->message_id=XMALLOC(lastchar-contentp+2);
					memcpy(meta->message_id,contentp,lastchar-contentp+1);
					meta->message_id[lastchar-contentp+1]='\0';
					continue;
				} else if (ml->next) {
					/* mehrere IDs? */
					meta->mausflags.exportflag=XR_MINUS;
					meta_add_trace(meta,MMT_ERROR|MMT_NX,
						"zu viele Message-IDs");
					meta_add_list (&meta->unknown_headers,
						"X-Mauscon-Original-Message-Id",
						contentp, 0);
					/* mail versuchen wir zu transportieren */
					meta->message_id=ml->text;
					ml->text=NULL;
					meta_free_list(&ml);
					if (!is_mail)
						errors_detected++;
				} else {
					meta->message_id=ml->text;
					ml->text=NULL;
					meta_free_list(&ml);
				}
			}
			else if (!strhalfcasecmp(namep,"ime-version"))
			{
				char *mv=rfc_getword(contentp,NULL);
				if (mv)
				{
					char *next;
					mime_version_major=strtol(mv,&next,10);
					if (next && *next++=='.')
						mime_version_minor=strtol(next,NULL,10);
					/* alles was > 1.0 ist ist vielleicht Gift fr uns,
					 * aber wir haben keinen Benutzer, den wir fragen
					 * knnten. Folglich versuchen wir >1.0 mglichst
					 * *wie* 1.0 zu behandeln. Hope it doesn't break
					 * anything.
					 */
					meta->mime_version=mv;
					if (mime_version_major!=1 
						|| mime_version_minor!=0) {
						meta->mausflags.exportflag=XR_MINUS;
							meta_add_trace(meta,MMT_ERROR|MMT_NX,
							"mime-version != 1.0 found: %s", mv);
					}
				}
				continue;
			}
			break;
		case 'n':
			/* Newsgroups? */
			if (!strhalfcasecmp(namep,"ewsgroups"))
			{
				if (!rfc_groups_to_metalist(&meta->newsgroups,contentp)) {
					meta->mausflags.exportflag=XR_MINUS;
					meta_add_trace(meta,MMT_ERROR|MMT_NX,
						"Newsgroups unwandelbar: %s", contentp);
					meta_add_list (&meta->unknown_headers,
						"X-Mauscon-Original-Newgroups",
						contentp, 0);
				}
				if (!meta->newsgroups && is_mail)
					errors_detected++; /* really can't handle this */
				continue;
			}
			break;
		case 'o':
			/* Organization */
			if (!strhalfcasecmp(namep,"rganization")) {
				meta->organization=XSTRDUP	(contentp);
				continue;
			}
			break;
		case 'p':
			/* Path */
			if (!strhalfcasecmp(namep,"ath")) {
				meta->path=XSTRDUP	(contentp);
				continue;
			}
			break;
		case 'r':
			/* references */
			if (!strhalfcasecmp(namep,"eferences")) {
				int troubles;
				if (!rfc_ids_to_metalist(&meta->references,contentp,&troubles)||troubles) {
					/* die ist defekt ... */
					meta->mausflags.exportflag=XR_MINUS;
					meta_add_trace(meta,MMT_ERROR|MMT_NX,
						"References unwandelbar: %s", contentp);
					meta_add_list (&meta->unknown_headers,
						"X-Mauscon-Original-References",
						contentp, 0);

				}
				continue;
			}
			/* Replaces, wird fr mail einfach ignoriert. */
			if (!strhalfcasecmp(namep,"eplaces") && !is_mail) {
				int troubles;
				if (!rfc_ids_to_metalist(&meta->replaces,contentp,&troubles) || troubles) {
					/* die ist defekt ... */
					meta->mausflags.exportflag=XR_MINUS;
					meta_add_trace(meta,MMT_ERROR|MMT_NX,
						"Replaces unwandelbar: %s", contentp);
					meta_add_list (&meta->unknown_headers,
						"X-Mauscon-Original-Replaces",
						contentp, 0);
					continue;
				}
				meta->mausflags.exportflag=XR_MINUS;
				meta_add_trace(meta,MMT_CONTROL|MMT_NX,
					"replaces found: %s", contentp);
			}
			/* Reply-To: */
			if (!strhalfcasecmp(namep,"eply-to")) {
				if (rfc_adress_to_metalist(1,&meta->reply_to,contentp)!=0 || !meta->reply_to) {
					meta->mausflags.exportflag=XR_MINUS;
					meta_add_trace(meta,MMT_ERROR|MMT_NX, "problems converting 'reply-to' header");
					meta_add_list (&meta->unknown_headers,
						"X-Mauscon-Original-Reply-To",
						contentp, 0);
				}
				continue;
			}
			/* return-path */
			if (!strhalfcasecmp(namep,"eturn-path") && !return_path) {
				metalist_t *ml=NULL;
				if (rfc_adress_to_metalist(1,&ml,contentp)!=0 || !ml) {
					meta_add_trace(meta,MMT_ERROR, 
						"problems converting 'return-path'");
					meta_add_list (&meta->unknown_headers,
						"X-Mauscon-Original-Return-Path",
						contentp, 0);
				} else {
					if (ml && ml->text)
						return_path=XSTRDUP(ml->text);
					if (ml->next) {
						meta_add_trace(meta,MMT_ERROR,
							"zu viele envelope-Absender: %s", contentp);
						meta_add_list (&meta->unknown_headers,
							"X-Mauscon-Original-Return-Path",
							contentp, 0);
					}
				}
				meta_free_list(&ml);
				continue;
			}
			break;
		case 's':
			if (*namep=='u') {
				/* Subject? */
				namep++;
				if (!strhalfcasecmp(namep,"bject")) {
					meta->subject=XSTRDUP(contentp);
					continue;
				} else if (!strhalfcasecmp(namep,"mmary")) {
					meta->summary=XSTRDUP(contentp);
					continue;
				} else if (!strhalfcasecmp(namep,"persedes") && !is_mail) {
					int troubles;
					if (!rfc_ids_to_metalist(&meta->supersedes,contentp,&troubles) || troubles) {
						/* die ist defekt ... */
						meta->mausflags.exportflag=XR_MINUS;
						meta_add_trace(meta,MMT_ERROR|MMT_NX,
							"Supersedes unwandelbar: %s", contentp);
						meta_add_list (&meta->unknown_headers,
							"X-Mauscon-Original-Supersedes",
							contentp, 0);
						continue;
					}
					/* fr mail einfach ignorieren */
					meta->mausflags.exportflag=XR_MINUS;
					meta_add_trace(meta,MMT_CONTROL|MMT_NX,
						"supersedes found: %s", contentp);
				}
			} else if (*namep=='e' && !strhalfcasecmp(namep+1,"nder")) {
				/* Es gibt zuviele kaputte Sender-Zeilen in der
				 * Welt, *seuufz*. Wenn ich mal sehr viel Zeit habe,
				 * schreibe ich einen Autoflamewerfer gegen die 
				 * Leute, die "@anywhere", den Realname oder soetwas
				 * da eintragen.
				 * Also nicht "meckern".
				 */
				if (rfc_adress_to_metalist(0,&meta->sender,contentp)!=0 || !meta->sender) {
					meta->mausflags.exportflag=XR_MINUS;
					meta_add_trace(meta,MMT_ERROR|MMT_NX, 
						"problems converting 'sender' header");
					meta_add_list (&meta->unknown_headers,
						"X-Mauscon-Original-Sender",
						contentp, 0);
				}
				continue;
			}
			break;
		case 't':
			/* To: ? */
			if (tolower(*namep)=='o' && namep[1]=='\0')
			{
				/* Wenn Mail, dann Fehler im Header ignorieren,
				 * wir haben ja den Envelope. Adressproblem 
				 * aber trotzdem ins Logfile eintragen.
				 */
				if (rfc_adress_to_metalist(1,&meta->head_to,contentp)!=0 || !meta->head_to) {
					meta->mausflags.exportflag=XR_MINUS;
					meta_add_trace(meta,MMT_ERROR|MMT_NX, 
						"problems converting 'to' header",
						contentp);
					meta_add_list (&meta->unknown_headers,
						"X-Mauscon-Original-To",
						contentp, 0);
				}
				continue;
			}
			break;
		case 'x':
			if (!strhalfcasecmp(namep,"-mausconnect-env-from")) {
				metalist_t *ml=NULL;
				if (rfc_adress_to_metalist(1,&ml,contentp)!=0 || !ml) {
					meta_add_trace(meta,MMT_ERROR, 
						"problems converting 'envelope-from'");
					meta_add_list (&meta->unknown_headers,
						"X-Mauscon-Original-Envelope-From",
						contentp, 0);
				} else {
					if (ml && ml->text)
						meta->env_from=XSTRDUP(ml->text);
					if (ml->next) {
						meta_add_trace(meta,MMT_ERROR,
							"zu viele envelope-Absender: %s", contentp);
						meta_add_list (&meta->unknown_headers,
							"X-Mauscon-Original-Envelope-From",
							contentp, 0);
					}
				}
				meta_free_list(&ml);
				continue;
			}
			else if (!strhalfcasecmp(namep,"-mausconnect-env-to"))
			{
				metalist_t *ml=NULL;
				if (rfc_adress_to_metalist(1,&ml,contentp)!=0 || !ml) {
					meta_add_trace(meta,MMT_ERROR, 
						"problems converting 'envelope-to'");
					meta_add_list (&meta->unknown_headers,
						"X-Mauscon-Original-Envelope-To",
						contentp, 0);
				} else {
					if (ml && ml->text) {
						if (opts.strip_bangpath)
							strip_bangpath(&(ml->text));
						meta->env_to=XSTRDUP(ml->text);
					}
					if (ml && ml->next) {
						meta_add_trace(meta,MMT_ERROR,
							"zu viele envelope-Empfaenger: %s", contentp);
						meta_add_list (&meta->unknown_headers,
							"X-Mauscon-Original-Envelope-To",
							contentp, 0);
					}
				}
				meta_free_list(&ml);
				continue;
			} else if (!strhalfcasecmp(namep,"-pgp-sig")) {
				/* PGP-Signierung in de.admin.news.announce */
				meta->mausflags.exportflag=XR_MINUS;
				meta_add_trace(meta,MMT_LOSS|MMT_NX,
					"x-pgp-sig found: %s", contentp);
			} else if (!strhalfcasecmp(namep,"-auth")) {
				/* PGP-Signierung in comp.os.linux.announce */
				meta->mausflags.exportflag=XR_MINUS;
				meta_add_trace(meta,MMT_LOSS|MMT_NX,
					"x-auth found: %s", contentp);
			}
			break;
		} /* switch */
	} /* while */

	if (!meta->env_to && delivered_to) 
		meta->env_to=delivered_to;
	else free(delivered_to);

	if (!meta->env_from && return_path) 
		meta->env_from=return_path;
	else free(return_path);

	if (is_mail && !meta->env_to) {
		/* seufz */
		/* f_maus.c knows how to deal with this and delivers the mail to 
		 * the "To:" header address - which is wrong, but still the
		 * best guess.
		 */
		meta_add_trace(meta,MMT_ERROR, 
			"no 'envelope-to' information found");
	}


	if (is_mail && !meta->env_from && opts.v_for_empty_envelope
		&& *opts.v_for_empty_envelope) {
		/* bounces */
		meta->env_from=XSTRDUP(opts.v_for_empty_envelope);
	}
	if (meta->env_from && !strchr(meta->env_from,'@'))
	{
		/* Smail will sich bei lokalen oder vermutliche lokalen
		 * Nachrichten nicht zum Erzeugen eines Domainparts 
		 * berreden lassen.
		 */
		char *p_tmp=meta->env_from;
		meta->env_from=XMALLOC(strlen(p_tmp)+2+strlen(opts.mysitedomain));
		sprintf(meta->env_from,"%s@%s",p_tmp,opts.mysitedomain);
		free(p_tmp);
	}
	if (is_mail && inrepto)
	{
		if (!meta->references)
			meta->references=inrepto;
		else
		{
			metalist_t *akt;
			akt=meta->references;
			while (akt->next)
				akt=akt->next;
			akt->next=inrepto;
		}
	}
	return errors_detected;
}

#define OUTPUT1(s1) do {fputs(s1,f); } while(0)
#define OUTPUT2(s1,s2) do {fputs(s1,f); fputs(s2,f);} while(0)
#define OUTPUT3(s1,s2,s3) do {fputs(s1,f); fputs(s2,f); fputs(s3,f);} while(0)
#define OUTPUT4(s1,s2,s3,s4) \
	do {fputs(s1,f); fputs(s2,f); fputs(s3,f); fputs(s4,f) ; }while(0)

void 
rfc_print_message(metamsg_t *meta,FILE *f, const char *crlf,int dotescape)
{
	uostr_t tmpstr=UOSTR_INIT;
	char datetmp[40];
	int strip_cr;

	if (!meta->is_mail)
	{
		if (meta->env_to)
			meta->is_mail=1;
		else
			meta->is_mail=0;
	}

	/* 
	 * Der Nachfolger von RFC 1036 sagt, Pflichtheader seien:
	 * Date, From, Message-ID, Subject, Newsgroups, Path.
	 * Alle die drfen nur einmal auftreten. In einigen Situationen
	 * sind auch References, Sender, Approved Pflicht.
	 * Sie sollten als erste in der Nachricht kommen, vor den 
	 * optionalen Headern und den nicht standardisierten (aus
	 * Effizienzgruenden).
	 */
	if (!meta->message_id && !meta->is_mail)
	{
		/* Haeh? */
		return;
	}

	/* paranoid, aber was soll's */
	if (meta->message_id)
	{
		char *id=entumlauten1(meta->message_id);
		OUTPUT4("Message-ID: <",id,">",crlf);
		free(id);
	}

	rfc_make_date(datetmp,sizeof(datetmp),meta->date);
	OUTPUT3("Date: ",datetmp,crlf);

	if (meta->head_from)
	{
		uostr_cut(&tmpstr,0);
		rfc_metalist2adrheader(meta->mime_code_headers,&tmpstr,meta->head_from);
		OUTPUT3("From: ",tmpstr.data,crlf);
	}
	else if (meta->env_from)
		OUTPUT3("From: ",meta->env_from,crlf);

	if (meta->subject)
	{
		int has_re=0;

		uostr_cut(&tmpstr,0);
		rfc_head27bit(meta->mime_code_headers,&tmpstr,meta->subject);
		uostr_x0(&tmpstr);
		if (!strnhalfcasecmp(tmpstr.data,"re: ",4))
			has_re=1;
		if (meta->references && !has_re)
			OUTPUT3("Subject: Re: ",tmpstr.data,crlf);
		else if (!meta->references && has_re)
			/* re: weg bei Nachrichten ohne References */
			OUTPUT3("Subject: ",tmpstr.data+4,crlf);
		else
			OUTPUT3("Subject: ",tmpstr.data,crlf);
	}
	else
		OUTPUT2("Subject: (empty)",crlf);

	if (meta->newsgroups)
	{
		metalist_t *p;
		p=meta->newsgroups;	
		if (meta->is_mail)
			fputs("X-Newsgroups: ",f);
		else
			fputs("Newsgroups: ",f);
		while (p)
		{
			fputs(p->text,f);
			if (p->next)
				fputs(",",f);
			p=p->next;
		}
		fputs(crlf,f);
	}

	/* Path ist in MAIL nicht sinnvoll, Received-Zeilen knnen wir
	 * nicht.
	 */
	if (meta->path && !meta->is_mail)
		OUTPUT3("Path: ",meta->path,crlf);

	if (meta->references && meta->is_mail)
	{
		char *entumlautet;
		metalist_t *p;

		p=meta->references;
		while (p->next)
			p=p->next;
		entumlautet=entumlauten1(p->text);
		XNULLCHECK(entumlautet);

		OUTPUT4("In-Reply-To: <",entumlautet,">",crlf);

		free(entumlautet);
	}
	else if (meta->references)
	{
		metalist_t *p;

		fputs("References:",f);

		p=meta->references;
		while (p)
		{
			if (p->text)
			{
				char *entumlautet=entumlauten1(p->text);
				XNULLCHECK(entumlautet);
				OUTPUT3(" <",entumlautet,">");
				free(entumlautet);
			}
			p=p->next;
		}
		fputs(crlf,f);
	}

	if (meta->sender)
	{
		rfc_metalist2adrheader(meta->mime_code_headers,&tmpstr,meta->sender);
		OUTPUT3("Sender: ",tmpstr.data,crlf);
	}


	/* Soweit die Pflichtheader. Nun die anderen Standardheader */

	if (meta->reply_to)
	{
		rfc_metalist2adrheader(meta->mime_code_headers,&tmpstr,meta->reply_to);
		OUTPUT3("Reply-To: ",tmpstr.data,crlf);
	}

	if (meta->followup_to)
	{
		metalist_t *p;
		p=meta->followup_to;	
		if (meta->is_mail)
			fputs("X-Followup-To: ",f);
		else
			fputs("Followup-To: ",f);
		while (p)
		{
			fputs(p->text,f);
			if (p->next)
				fputs(",",f);
			p=p->next;
		}
		fputs(crlf,f);
	}

#define SIMPLE_LINE(test,var,tag) \
	if (test && meta->var) { \
		uostr_cut(&tmpstr,0); \
		rfc_head27bit(meta->mime_code_headers,&tmpstr,meta->var); \
		uostr_x0(&tmpstr); \
		OUTPUT3(tag ": ",tmpstr.data,crlf); \
	}
#define ID_LINE(test,var,tag) do {\
	if (test && meta->var && meta->var->text) { \
		int tcount=0; size_t tlen=0; \
		metalist_t *mlp=meta->var; \
		OUTPUT1(tag ":"); tlen=strlen(tag+2); while (mlp) { \
			size_t l1=strlen(mlp->text); if (l1+tlen>78 && tcount) { \
				OUTPUT2("\n ",mlp->text); tcount=1; tlen=l1+2; \
			} else { \
				OUTPUT2(" ",mlp->text); tcount++; tlen+=l1+1; \
			} \
		} \
		OUTPUT1(crlf); \
	} } while(0)

	SIMPLE_LINE(1,keywords,"Keywords")
	SIMPLE_LINE(1,summary,"Summary")
	SIMPLE_LINE(1,organization,"Organization")
	SIMPLE_LINE(!meta->is_mail,approved,"Approved")
	SIMPLE_LINE(!meta->is_mail,distribution,"Distribution")

	/* MIME-Header erzeugen, falls dies ntig ist. Was wir hier vllig
	 * ignorieren ist die Mglichkeit, da der Server, der die Mails
	 * als nchster bekommt, 8bit nicht mag und bounct!
	 */
	if (meta->mime_version)
	{
		OUTPUT3("MIME-Version: ",meta->mime_version,crlf);
		if (meta->content_type)
			OUTPUT3("Content-Type: ",meta->content_type,crlf);
		else
			OUTPUT2("Content-Type: text/plain; charset=ISO-8859-1",crlf);
		if (meta->content_transfer_encoding)
			OUTPUT3("Content-Transfer-Encoding: ",
					meta->content_transfer_encoding,crlf);
		else if (meta->bodystat.achtbit && meta->mime_code_body)
			OUTPUT2("Content-Transfer-Encoding: quoted-printable",crlf);
		else if (meta->bodystat.achtbit)
			OUTPUT2("Content-Transfer-Encoding: 8bit",crlf);
		if (meta->content_description)
			OUTPUT3("Content-Description: ",meta->content_description,crlf);
		if (meta->content_id)
			OUTPUT3("Content-ID: ",meta->content_id,crlf);
	}
	else if (meta->bodystat.achtbit && opts.mime_headerlines)
	{
		OUTPUT2("MIME-Version: 1.0",crlf);
		OUTPUT2("Content-Type: text/plain; charset=ISO-8859-1",crlf);
		if (meta->mime_code_body)
			OUTPUT2("Content-Transfer-Encoding: quoted-printable",crlf);
		else
			OUTPUT2("Content-Transfer-Encoding: 8bit",crlf);
	}

    if (meta->head_to && !meta->x_to)
	{
		/* Der 1036-Nachfolger sagt, Point-to-Point-Header wie
		 * To und Cc gehren nicht in News. 
		 */
		rfc_metalist2adrheader(meta->mime_code_headers,&tmpstr,meta->head_to);
		if (meta->is_mail)
			OUTPUT3("To: ",tmpstr.data,crlf);
		else
			OUTPUT3("X-To: ",tmpstr.data,crlf);
	}

	SIMPLE_LINE(1,x_to,"X-To");

	if (opts.workmode!=USER) {
		fputs("X-Gateway: Mauscon(" VERSION ")/",f);
		fputs(opts.mysitedomain,f);
		if (meta->x_format)
		{
			OUTPUT2(", ",meta->x_format);
		}
		fputs(crlf,f);
		fputs("X-Gateway-Admin: ",f);
		fputs(opts.gatemaster,f);
		fputs(crlf,f);
	} else {
		fputs("X-Gateway: Mauscon(" VERSION ")/",f);
		fputs(opts.mysitedomain,f);
		if (meta->x_format)
		{
			OUTPUT2(", ",meta->x_format);
		}
		fputs(", admin: ",f);
		fputs(opts.gatemaster,f);
		fputs(crlf,f);
	}

	SIMPLE_LINE(1,control,"Control");
	SIMPLE_LINE(1,also_control,"Also-Control");
	ID_LINE(1,supersedes,"Supersedes");

	if (meta->trace_header) {
		metalist_t *akt;
		akt=meta->trace_header;
		while (akt) {
			unsigned long n=strtoul(akt->text,0,10);
			if (n & opts.debugmask_rfc) {
				fputs("X-Mauscon-Trace: ",f);
				if (n & MMT_ERROR)
					fputs("Error: ",f);
				OUTPUT2(akt->ext,crlf);
			}
			akt=akt->next;
		}
	}
	if (meta->unknown_headers)
	{
		metalist_t *akt;
		akt=meta->unknown_headers;
		while(akt)
		{
			fputs(akt->text,f);
			/* Sonderbehandlung fr Fortsetzungszeilen */
			if (!isspace(akt->text[0])) {
				fputs(": ",f);
			}
			if (akt->ext)
				fputs(akt->ext,f);
			fputs(crlf,f);
			akt=akt->next;
		}
	}

	/* Leerzeile zum Abschlu des Headers schreiben */
	fputs(crlf,f);

	/* Jetzt wird es interessanter */

	/* strip_cr: wenn FALSE knnen wir die Nachricht ohne CR/LF nach LF 
	 * zu wandeln schreiben (NNTP versus rnews)
	 */
	strip_cr=strcmp(crlf,"\r\n");

	if (strip_cr || (meta->mime_code_body && meta->bodystat.achtbit))	
	{
		char *start;
		char *p;
		char saved='\0'; /* keep gcc happy */
		start=meta->body.data;
		while (start && *start)
		{
			p=strchr(start,'\n');
			if (p)
			{
				if (p>start && p[-1]=='\r')
					p--;
				saved=*p;
				*p='\0';
			}
			if (*start=='.' && dotescape)
			/* printf ist je nach C-Library recht teuer */
				fputc('.',f);

			if (meta->mime_code_body) {
				/* so, jetzt wird es haesslich ... */
				unsigned char *q=start;
				const char hex[]="0123456789ABCDEF";
				int cnt=0;
				while (*q) {
					size_t l=strspn(q,
"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
"!#$%&'()*+,-./:;<>?@[\\]^_`{|}~\"\t ");
					if (l+cnt>76) {
						l=76-cnt;
						while (l>0 && !isspace(q[l]))
							l--;
					}
					if (l==0) {
						if (cnt <= 73) {
							fputc('=',f);
							fputc(hex[(*q) / 16],f);
							fputc(hex[(*q) & 0x0f],f);
							q++;
							cnt+=3;
						} else {
							/* soft linebreak */
							fputc('=',f);
							fputs(crlf,f);
							cnt=0;
						}
					} else if (l==1) {
						fputc(*q,f);
						cnt++;
						q++;
					} else {
						fwrite(q,1,l,f);
						cnt+=l;
						q+=l;
					}
					if (cnt>=76) {
						fputc('=',f);
						fputs(crlf,f);
						cnt=0;
					}
				}
				if (' '==q[-1] || '\t'==q[-1]) {
					/* rule3: whitespace at end of line */
					fputc('=',f);
					fputs(crlf,f);
				}
				/*fputs("=0D=0A",f);*/ /* rule 4 */
				fputs(crlf,f);
			} else {
				fputs(start,f);
				fputs(crlf,f);
			}
			if (p)
			{
				*p=saved;
				if (saved=='\r')
					start=p+2;
				else
					start=p+1;
			}
			else
			{
				fputs(crlf,f);
				start=NULL;
			}
		}
	}
	else
	{
		char *start;
		char *p;
		size_t l;
		start=meta->body.data;
		while (start && *start)
		{
			BOOLEAN need_crlf=FALSE;
			p=strchr(start,'\n');
			if (!p)
				l=strlen(start);
			else
				l=p-start+1;
			if (*start=='.' && dotescape)
				fputc('.',f);
			if (p) {
				/* mglicherweise kommen hier Zeilen ohne \r an. rfc_to_meta
				 * tut das beispielsweise. */
				if (p==start) {
					fputc('\r',f);
				} else if (p[-1]!='\r') {
					l-=1;
					need_crlf=TRUE;
				}
			}
			fwrite(start,l,1,f);
			if (p) {
				start=p+1;
				if (need_crlf)
					fputs(crlf,f);
			} else {
				fputs(crlf,f);
				start=NULL;
			}
		}
	}
	uostr_freedata(&tmpstr);
}

