/* Mauscon: Nachrichtenformatskonverter fuer das MausNet
   Copyright (C) 1992 1993 1994 1995 Michael Kraehe, Uwe Ohse

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   rfc2maus.c: Konvertierung von RFC nach MausTausch
*/

#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <unistd.h>

#include <errno.h>
#include <ctype.h>
#include <fcntl.h>
#include <syslog.h>

#if HAVE_DIRENT_H
# include <dirent.h>
# define NAMLEN(dirent) strlen((dirent)->d_name)
#else
# define dirent direct
# define NAMLEN(dirent) (dirent)->d_namlen
# if HAVE_SYS_NDIR_H
#  include <sys/ndir.h>
# endif
# if HAVE_SYS_DIR_H
#  include <sys/dir.h>
# endif
# if HAVE_NDIR_H
#  include <ndir.h>
# endif
#endif

#include <sys/types.h>

#ifdef HAVE_SYS_STAT_H
#  include <sys/stat.h>
#endif

#ifdef HAVE_MMAP
#include <sys/mman.h>
#ifndef MAP_FILE
#define MAP_FILE 0
#endif
#endif

#include "mausconlib.h"
#include "mauscon.h"
#include "metamsg.h"
#include "errormail.h"
#include "rfc.h"
#include "maus.h"
#include "groups.h"
#include "domains.h"
#include "rfcaddr.h"
#include "msg_io.h"
#include "uostr.h"
#include "dynhash.h"
#include "crc32.h"
#include "uogetopt.h"
#include "setvars.h"

/* Weil MAUS \r zwingend braucht */
#define LINEEND "\r\n"


int opt_only_convert;
static int opt_unlink_news;
static int opt_unlink_appended_news;
int work_only;
int dont_use_nntp;
int dont_use_smtp;
int opt_quark_mail2spool;
static char *opt_outputfilename;
static int opt_read_header_for_envelope=0;


static void write_cmd(FILE *outputfile);
static BOOLEAN rfc2maus(mio_t *ochan, metamsg_t *meta, char *unlink_fname);
static void convert_news (void);
static void convert_mail (mio_t *ochan);
static BOOLEAN convert_one_file (mio_t *ochan, char *fname, uostr_t *mbuf, metamsg_t *meta, BOOLEAN is_mail);
static void	append_news(FILE *outputfile);
static void quark_mail2spool(const char *user,time_t since);
static void quark_mail2spool_oneuser(const char *dir,time_t since,uostr_t *mbuf,metamsg_t *meta);
static FILE * quark_mail2spool_loaddb(const char *dir, uostr_t *dbfname1, uostr_t *dbfname2, void **handle);
static void quark_mail2spool_savedb(uostr_t *dbfname1, uostr_t *dbfname2,void *handle, FILE *f);

static void
diropen_error(const char *dirname)
{
	uostr_t m=UOSTR_INIT;
	uostr_t g=UOSTR_INIT;
	uostr_xadd_cstrmulti(&m,dirname,": ",strerror(errno),"\n",NULL);
	uostr_x0(&m);
	uostr_xadd_cstrmulti(&m,"-",opts.gatemaster,NULL);
	uostr_x0(&g);
	errormail (NULL, "error creating infile", m.data, g.data, NULL);
	uostr_freedata(&m);
	uostr_freedata(&g);
}

static BOOLEAN
rfc2maus(mio_t *ochan, metamsg_t *msg, char *unlink_fname)
{
	cancel_t *cancel;
	unsigned const char *error_text = NULL;
	int errors_detected;

	errors_detected = 0;
	error_text=NULL;

	if (msg->message_id && !msg->is_mail && opts.honor_cancels & HONOR_CANCEL_SELF)
	{
		cancel=cancel_find_in_list(msg->message_id);
		if (cancel 
			&& cancel->from && msg->head_from && msg->head_from->text
			&& !strcasecmp(msg->head_from->text,cancel->from))
		{
			/* Eine Cancel-Nachricht auf die zu verarbeitende Nachricht ist 
			 * also schon eingetroffen. 
			 */		
			msglog("%s: canceled\n",msg->message_id);
			if (unlink_fname)
				unlink(unlink_fname);
			cancel->done=1;
			return TRUE;
		}
	}
	if (!msg->is_mail)
	{
		BOOLEAN is_control=FALSE;
		if (msg->supersedes && msg->head_from && msg->head_from->text && opts.honor_cancels) {
			metalist_t *mlp=NULL;
			mlp=msg->supersedes;
			while (mlp) {
				cancel_add(mlp->text,msg->head_from->text);
				mlp=mlp->next;
			}
		} 
		if (msg->replaces && msg->head_from && msg->head_from->text && opts.honor_cancels) {
			/* XXX this is not the perfect way to handle Replaces: */
			metalist_t *mlp=NULL;
			mlp=msg->replaces;
			while (mlp) {
				cancel_add(mlp->text,msg->head_from->text);
				mlp=mlp->next;
			}
		} 
		if (msg->control && msg->head_from && msg->head_from->text && opts.honor_cancels) {
			if (!strncmp(msg->control,"cancel ",7)) {
				metalist_t *ml=NULL;
				metalist_t *mlp=NULL;
				int troubles;
				if (rfc_ids_to_metalist(&ml,msg->control+7,&troubles) && !troubles) {
					mlp=ml;
					while (mlp) {
						cancel_add(mlp->text,msg->head_from->text);
						mlp=mlp->next;
					}
				}
				meta_free_list(&ml);
			}
			is_control=TRUE;
		}
		if (msg->also_control && msg->head_from && msg->head_from->text && opts.honor_cancels) {
			if (!strncmp(msg->also_control,"cancel ",7)) {
				metalist_t *ml=NULL;
				metalist_t *mlp=NULL;
				int troubles;
				if (rfc_ids_to_metalist(&ml,msg->also_control+7,&troubles) && !troubles) {
					mlp=ml;
					while (mlp) {
						cancel_add(mlp->text,msg->head_from->text);
						mlp=mlp->next;
					}
				}
				meta_free_list(&ml);
			}
		}
		if (is_control)
			return TRUE;
	}

	/*
	 * Der Header ist in die Datenstrukturen konvertiert worden.
	 * Nun Vollstndigkeit prfen. Allzu viel ist nicht zu tun.
	 * - bei Mail wird die Angabe der Envelope-Adressen geprft, und
	 *   geprft, ob der Envelope-Absender durch das Gate schreiben darf.
	 * - bei News wird geprft, ob eine der Gruppen bekannt ist, ob
	 *   From oder Sender durch das Gate schreiben darf.
	 * Diese Prfungen mssen wohl beim Boxmodus anders gehandhabt werden
	 * (eine Nachricht, die bis hier kommt, ist als OK anzusehen?)
	 */
	 if (msg->is_mail)
	 {
	 	if (!opt_only_convert)
	 	{
			if (!msg->env_to && opt_read_header_for_envelope) {
				int count=0;
				/* Uhh ... igitt. */
				metalist_t *a;
				for (a=msg->head_to;a&&a->text;a=a->next) {
					msg->env_to=XSTRDUP(a->text);
					if (rfc2maus(ochan, msg, unlink_fname))
						count++;
					free(msg->env_to);
					msg->env_to=NULL;
				}
				for (a=msg->cc;a&&a->text;a=a->next) {
					msg->env_to=XSTRDUP(a->text);
					if (rfc2maus(ochan, msg, unlink_fname))
						count++;
					free(msg->env_to);
					msg->env_to=NULL;
				}
				for (a=msg->bcc;a&&a->text;a=a->next) {
					msg->env_to=XSTRDUP(a->text);
					if (rfc2maus(ochan, msg, unlink_fname))
						count++;
					free(msg->env_to);
					msg->env_to=NULL;
				}
				if (count)
					return TRUE;
			}
			if (!msg->env_from && opt_read_header_for_envelope) {
				if (msg->head_from)
					msg->env_from=XSTRDUP(msg->head_from->text);
			}

			if (!msg->env_to)
			{
				errors_detected++;
				error_text = "no receiver in envelope";
			}

			if (!msg->env_from)
			{
				errors_detected++;
				error_text = "no sender in envelope";
			}
			else if (!is_allowed(msg->env_from))
			{
				errors_detected++;
				error_text = "envelope-sender is not allowed to use the gateway";
			}
		}
	 }
	 else
	 {
	 	/* News */
	 	if (msg->newsgroups)
	 	{
	 		metalist_t *ng;
	 		BOOLEAN found=FALSE;
			IMPORT_STATE import_state=IMPORT_SKIP;
	 		ng=msg->newsgroups;
	 		while(ng)
 			{
	 			grp_data_t *grp;
				grp = grp_suche_rfcname (ng->text,opts.pollsite);
				if (grp)
				{
					found=TRUE;
					if (grp->import_state > import_state)
						import_state=grp->import_state;
					if (grp->mausnet_distribution)
					{
						BOOLEAN do_it=TRUE;
						if (msg->distribution)
						{
							/* Wenn die Distribution nicht schon lokal ist */
							if (strcasecmp(msg->distribution, 
									opts.usenet_dist_lokal))
								free(msg->distribution);
							else
								do_it=FALSE;
						}
						if (do_it)
							msg->distribution=
								XSTRDUP(grp->mausnet_distribution);
					}
				}
 				ng=ng->next;
	 		}
			if (import_state==IMPORT_KILL && opts.workmode==GATE)
			{
				errors_detected++;
				error_text = _("import into at least one newsgroup not allowed");
			}
			if (import_state==IMPORT_SKIP && opts.workmode==GATE && found)
			{
				errors_detected++;
				error_text = _("all newsgroup skipped");
			}
	 		if (!found)
	 		{
				/* 
				 * Dummyeintrag in der Namensliste erzeugen, damit ber
				 * einen Gruppenamen nur einmal gemeckert wird.
				 */
	 			grp_data_t *grp;
				grp=XMALLOC(sizeof(grp_data_t));
				*grp=leer_grp;
				grp->mausname=XSTRDUP(msg->newsgroups->text);
				grp->rfcname=XSTRDUP(msg->newsgroups->text);
				grp->boxen=XSTRDUP(opts.pollsite);
				grp->dont_save=1;
				grp->import_state=IMPORT_OK;
				grp_append_new(grp);
			}
	 	}
	 	else
	 	{
			errors_detected++;
			error_text = _("no newsgroup given");
	 	}

	 	if (msg->date==0 || msg->date==(uo_sec70_t) -1) {
			errors_detected++;
			error_text = _("date not given or date unparsable");
	 	}

	 	if (!msg->head_from || !msg->head_from->text) {
	 		/* Keine From-Zeile? Laut 1036 verboten */
			errors_detected++;
			error_text = _("from not given");
	 	}
	 	else
	 	{
	 		/* news haben laut 1036-Nachfolger nur einen Sender/From */
	 		int ok=0;
	 		if (is_allowed(msg->head_from->text))
	 			ok=1;
	 		else if (msg->sender && msg->sender->text)
	 		{
	 			if (is_allowed(msg->sender->text))
	 				ok=1;
				else
					error_text = _("from and sender address are not allowed to use the gateway");
	 		}
			else
				error_text = _("from address is not allowed to use the gateway");
	 		if (!ok)
	 		{
				errors_detected++;
				if (opts.unlink_unallowed && unlink_fname)
					unlink(unlink_fname);
			}
		}
	}
	if (!msg->message_id)
	{
		if (!msg->is_mail)
		{
			errors_detected++;
			error_text = _("no message-id given");
		}
		else
		{
			char numbuf[STR_ULONG];
			uostr_t idbuf=UOSTR_INIT;
			static int counter=0;
			uostr_xadd_cstr(&idbuf,"No.Id.Given.");
			str_ulong(numbuf,getpid());
			uostr_xadd_cstr(&idbuf,numbuf);
			uostr_xadd_char(&idbuf,'.');
			str_ulong(numbuf,++counter);
			uostr_xadd_cstr(&idbuf,numbuf);
			uostr_xadd_char(&idbuf,'@');
			uostr_xadd_cstr(&idbuf,opts.mysitedomain);
			uostr_x0(&idbuf);
			msg->message_id=XSTRDUP(idbuf.data);
			msglog("[no msgid given -> %s]\n",msg->message_id);
			uostr_freedata(&idbuf);
		}
	}
	if (opts.workmode==USER && !msg->is_mail)
	{
		/* Wiedereinleitungen von Nachrichten vermeiden - das *sollte*
		 * zwar auch durch die Absenderkontrolle abgefangen werden,
		 * aber da es so leicht ist, es Newssystem zu verkonfigurieren,
		 * ist diese Paranoia zwecks Dupevermeidung sinnvoll.
		 */
		if (msg->path && strstr(opts.pollsite,msg->path))
		{
			errors_detected++;
			error_text = _("reimport stopped");
		}
	}

	if (errors_detected != 0)
	{
		/* Ob stderr der richtige Platz ist??? */
		if (msg->message_id)
			fprintf (stderr, "ID:%s\n", msg->message_id);

		fprintf (stderr, _("%d errors detected\n"), errors_detected);
		fprintf (stderr, _("one of them: %s\n"), error_text);
		return FALSE;
	}

	/***********************************************************************
	*
	*                    Nun Ausgabe im MausTausch-Format
	*
	************************************************************************/
	ochan->fn->export(ochan,msg);

	return TRUE;
}

static	const char *opt_cnffname;
static	char *opt_single_mail_fname=NULL;
static	char *opt_single_news_fname=NULL;
static uogetopt_t myopts[]={
    {'a',"action",UOGO_STRING|UOGO_HIDDEN,&action,0,
    "Channel to use (old name)","CHANNEL"},
    {'C',"channel",UOGO_STRING,&action,0,
    "Channel to use","CHANNEL"},
    {'c',"cnffile",UOGO_STRING,&opt_cnffname,0,
    "Configuration file to use","FILE"},
    {'c',"cnf-file",UOGO_STRING|UOGO_HIDDEN,&opt_cnffname,0,
    "Configuration file to use","FILE"},
	{  0,"dont-use-nntp", UOGO_FLAG, &dont_use_nntp, 1,
	"do not send messages using NNTP", NULL},
	{  0,"dont-use-smtp", UOGO_FLAG, &dont_use_smtp, 1,
	"do not send messages using SMTP", NULL},
	{'h', "read-header", UOGO_FLAG,&opt_read_header_for_envelope,1,
	"read mail headers for envelope information.\n"
	/*------------------------------------------------*/
	"If this option is used and envelope information\n"
	"is missing then rfc2maus searches the To, Cc and\n"
	"Bcc headers and sends the message to all of those\n"
	"addresses.\n"
	"Also the From header will be used in case there\n"
	"is no envelope sender information.\n"
	"\nTranslation: DO NOT USE THIS OPTION\n"
	"unless you really know what you are doing.\n",NULL},

    {'M',"input-mail",UOGO_STRING,&opt_single_mail_fname,0,
    "read a single mail from FILE\n", "FILE"},
    {'N',"input-news",UOGO_STRING,&opt_single_news_fname,0,
    "read a single news from FILE\n", "FILE"},
	{'O',"only-convert", UOGO_FLAG, &opt_only_convert, 1,
	"don't send anything, but send to stdout\n"
	"use for testing, not for real work!", NULL},
	{  0,"only-mail", UOGO_FLAG, &work_only, ONLY_MAIL,
	"convert only mail\nWill not touch news or special things",NULL},
	{  0,"only-news", UOGO_FLAG, &work_only, ONLY_NEWS,
	"convert only news\nWill not touch mail or special things",NULL},
	{  0,"only-special", UOGO_FLAG, &work_only, ONLY_SPECIAL,
	"convert only specials\nWill not touch mail or news",NULL},
    {'o',"output",UOGO_STRING,&opt_outputfilename,0,
    "Save output to FILE\n"
	"otherwise it will be written to stdout","FILE"},
	{  0,"quark-mail2spool", UOGO_FLAG, &opt_quark_mail2spool, 1,
	"do not use this\nprivate playing ground for uo", NULL},
    {'s',"set",UOGO_CALLBACK,&setvarfn,0,
    "override variables in the configuration file","VAR=VALUE"},
	{  0,"unlink-news", UOGO_FLAG, &opt_unlink_news, 1,
	"throw newsbatches away after conversion.\n"
	"This is meant for user mode, not gateway mode",NULL},
	{  0,"unlink-appended_news", UOGO_FLAG, &opt_unlink_appended_news, 1,
	"throw newsbatches away after printing them.\n"
	"This is meant for user mode, not gateway mode",NULL},

    { 0 ,"copyright", UOGO_PRINT|UOGO_HIDDEN,0,0,
    "Copyright (C) 1999 Uwe Ohse\n"
    PACKAGE " comes with ABSOLUTELY NO WARRANTY,\n"
    "to the extent permitted by law.\n"
    "You may redistribute copies of " PACKAGE "\n"
    "under the terms of the GNU General Public License.\n"
    "For more information about these matters, see the file named COPYING.\n"
    "You may redistribute copies of " PACKAGE "\n",0},

    {0,0,0,0,0,0,0} /* --help and --version */
};


int
main (int argc, char **argv)
{
	mio_t *ochan;
	uostr_t errbuf=UOSTR_INIT;
	int exitcode;

	const char *slash;

	program_name=argv[0];
	slash=strrchr(program_name,'/');
	if (slash)
		program_name=slash+1;

    uogetopt (program_name, PACKAGE, VERSION, &argc, argv, uogetopt_out,
      "usage: rfc2maus [options] [channelname]\n"
      "convert RFC messages to MausTausch\n",
      myopts,
      "\nReport bugs to <uwe-mauscon@bulkmail.ohse.de>");

	if (!opt_cnffname)
		opt_cnffname=SYSCONFDIR "/" CNFNAME;

	if (!action) {
		if (!argv[1]) {
			log (LOG_ERR, _("channel not given"));
			exit(2);
		}
		action=argv[1];
		argv++;
	}
	if (argv[1]) {
		log (LOG_ERR, _("too many arguments"));
		exit (2);
	}

	/* configfile lesen */
    if (do_conf (opt_cnffname, action,
		setvarscount ? setvars : NULL,setvarscount))
        exit (1);

	openlog (program_name, 0, opts.sl_facility);
	setlogmask(LOG_UPTO(opts.sl_priority));
	open_msglog();

	if (!grp_load_list(FALSE))
		exit (1);				/* Fehlermeldung schon ausgegeben */

	if (!dom_init())
		exit(1);

	if (opt_quark_mail2spool) {
		quark_mail2spool(NULL,0);
		exit (0);
	}
	/* Cancel-Liste ffnen und Lock halten */
	if (cancel_init (1))
		exit (1);				/* Fehlermeldung schon ausgegeben */

	if (MIO_OK!=mio_maustausch.open(
		opt_outputfilename ? opt_outputfilename : "-", 
		"w",
		&ochan,
		&errbuf))
	{
		fprintf (stderr, "rfc2maus: %s: %s\n", 
			opt_outputfilename ? opt_outputfilename : "stdin",
			 strerror (errno));
		exit (1);
	}
	if (opt_single_mail_fname) {
		uostr_t mbuf=UOSTR_INIT;
		metamsg_t *meta=meta_alloc();
		XNULLCHECK(meta);
		exitcode=!convert_one_file (ochan, opt_single_mail_fname, &mbuf, meta, TRUE);
		uostr_freedata(&errbuf);
		ochan->fn->close(ochan);
	} else if (opt_single_news_fname) {
		uostr_t mbuf=UOSTR_INIT;
		metamsg_t *meta=meta_alloc();
		XNULLCHECK(meta);
		exitcode=!convert_one_file (ochan, opt_single_news_fname, &mbuf, meta, FALSE);
		uostr_freedata(&errbuf);
		ochan->fn->close(ochan);
	} else {
		uostr_freedata(&errbuf);
		convert_mail (ochan);
		ochan->fn->close(ochan);

		convert_news ();

		if (opt_outputfilename)
		{
			FILE *outputfile=fopen(opt_outputfilename,"a");
			if (!outputfile)
			{
				fprintf (stderr, _("rfc2maus: cannot fopen %s: %s\n"), 
					opt_outputfilename, strerror (errno));
			}
			append_news(outputfile);
			write_cmd(outputfile);
			fprintf(outputfile,"#"LINEEND);
		}
		else
		{
			append_news(stdout);
			write_cmd(stdout);
			fprintf(stdout,"#"LINEEND);
		}
		exitcode=0;
	}

	/* Alle Cancels lschen - die sind jetzt verarbeitet */
	cancel_end(1,0);

	closelog ();
	close_msglog();
        
#ifdef LEAK_CHECK
	GC_gcollect();
#endif
	exit (exitcode);
}

static void 
write_cmd(FILE *outputfile)
{
	uostr_t fname=UOSTR_INIT;
	uostr_t lockfname=UOSTR_INIT;
	int fatal_err=0;
	char str[1024];
	char *p;
	FILE *f;
	minfo_t *akt;
	cancel_t *cakt;
	BOOLEAN cmd_written=FALSE;

	uostr_xdup_cstrmulti(&lockfname,opts.spoolbase,"/",action,"/etc/LOCK",NULL);
	uostr_x0(&lockfname);
	while (lock_create (lockfname.data, &fatal_err))
	{
		if (fatal_err)
		{
			log(LOG_ERR, _("cannot lock with %s"),lockfname.data);
			return;
		}
		sleep(5);
	}

	/* 
	 * Eigentlich mte man jetzt die noch brigen Cancels verarbeiten,
	 * sie irgendwie in die MausNetinternen IDs umsetzen ... 
	 * Brute Hack Alert.
	 */
	cakt=cancel_anker();
	if (cakt && opts.honor_cancels & HONOR_CANCEL_PASS) {
		uostr_xdup_cstrmulti(&fname,opts.spoolbase,"/",action,"/etc/status.msg",NULL);
		uostr_x0(&fname);
		f=fopen(fname.data,"a");
		if (f) {
			while(cakt) {
				if (!cakt->done) {
					if (opts.workmode != USER 
						|| strstr(cakt->id,opts.mausdomain))
					{
						fprintf(f,"#%s\n",cakt->id);
						if (cakt->from && opts.cancel_bx_with_v)
							fprintf(f,"V%s\n",cakt->from);
						fprintf(f,"BX\n");
					}
				}
				cakt=cakt->next;
			}
			fclose(f);
		}
	}

	if (0==info_read_list())
	{
		akt=info_get_anker();
		fprintf(outputfile,"#CMD" LINEEND);
		cmd_written=TRUE;
		while (akt)
		{
			if (akt->needs_crc && akt->subscribed)
			{
				if (time(NULL)>akt->last_date+akt->pause*86400)
					fprintf(outputfile,":%s %ld"LINEEND,akt->kuerzel,
							akt->last_crc);
			}
			else if (akt->subscribed)
			{
				if (time(NULL)>akt->last_date+akt->pause*86400)
					fprintf(outputfile,":%s"LINEEND,akt->kuerzel);
			}
			akt=akt->next;
		}
	}

	uostr_xdup_cstrmulti(&fname,opts.spoolbase,"/",action,"/etc/infile.cmd",NULL);
	uostr_x0(&fname);
	f=fopen(fname.data,"r");
	if (f)
	{
		if (!cmd_written)
		{
			fprintf(outputfile,"#CMD" LINEEND);
			cmd_written=TRUE;
		}
		while (fgets (str, sizeof(str), f))
		{
			p=strpbrk(str,"\r\n");
			if (p)
				*p='\0';
			if (*str==':')
				fprintf(outputfile,"%s"LINEEND,str);
			else
				fprintf(outputfile,":%s"LINEEND,str);
		}
		fclose(f);
	}

	uostr_xdup_cstrmulti(&fname,opts.spoolbase,"/",action,"/etc/once.cmd",NULL);
	uostr_x0(&fname);
	f=fopen(fname.data,"r");
	if (f)
	{
		if (!cmd_written)
			fprintf(outputfile,"#CMD" LINEEND);
		while (fgets (str, sizeof(str), f))
		{
			p=strpbrk(str,"\r\n");
			if (p)
				*p='\0';
			if (*p=='#' || !*p) 
				continue;
			if (*str==':')
				fprintf(outputfile,"%s"LINEEND,str);
			else
				fprintf(outputfile,":%s"LINEEND,str);
		}
		fclose(f);
		log(LOG_INFO,_("added once.cmd"));
		unlink(fname.data);
	}
	uostr_xdup_cstrmulti(&fname,opts.spoolbase,"/",action,"/etc/status.msg",NULL);
	uostr_x0(&fname);
	f=fopen(fname.data,"r");
	if (f)
	{	
		unlink(fname.data);
		while (fgets (str, sizeof(str), f))
		{
			p=strpbrk(str,"\r\n");
			if (p)
				*p='\0';
			fprintf(outputfile,"%s"LINEEND,str);
		}
		fclose(f);
	}
	lock_delete(lockfname.data);
	uostr_freedata(&lockfname);
	uostr_freedata(&fname);
}

#ifndef O_BINARY
#define O_BINARY 0
#endif

static BOOLEAN
append_file(const char *fname, FILE *out)
{
    int fdin = -1;
	struct stat st;
	int save_errno=0;
	caddr_t src_addr=(caddr_t) -1;
	BOOLEAN ok;
	ssize_t slaenge;
	
	fdin = open(fname, O_RDONLY | O_BINARY);
	if (fdin < 0) {
		return FALSE;
	}
	if (fstat(fdin,&st))
	{
		save_errno=errno;
		close(fdin);
		errno=save_errno;
		return FALSE;
	}
	slaenge=(ssize_t) st.st_size;
#ifdef HAVE_MMAP
	src_addr = mmap(0, slaenge, PROT_READ,
					MAP_FILE | MAP_SHARED, fdin, 0);
	if (src_addr==(caddr_t)-1)
	{
		save_errno=errno;
		goto no_mmap;
	}
	/* mmap ok */
	close(fdin);
	
	if (slaenge!=(ssize_t) fwrite(src_addr,1,slaenge,out))
	{
		save_errno=errno;
		ok=FALSE;
	}
	else
		ok=TRUE;
	munmap(src_addr,slaenge);
	errno=save_errno;
	return ok;
no_mmap:
#endif
#define COPYBUFFER (1024 * 32)
	{
        char *buffer;
		ssize_t got;
		ssize_t written=0;
		
		buffer = (char *) malloc(COPYBUFFER);
		if (buffer == NULL)
		{
			close(fdin);
			errno=ENOMEM;
			return FALSE;
		}

		while ((got=xread(fdin,buffer,
						  COPYBUFFER > (size_t) slaenge ? slaenge : COPYBUFFER))>0)
		{
			slaenge-=got;
			/* it's beyond me: why do fwrite/fread return size_t, and read/write return
			 * ssize_t? 
			 */
			if (fwrite(buffer,1,got,out)!=(size_t)got) {
				save_errno=errno;
				break;
			}
		}
		if (got==-1)
			save_errno=errno;
		free(buffer);
		close(fdin);
		
		errno=save_errno;
		if (got==-1 || written==-1)
			return FALSE;
		return TRUE;
	}
}

static void
append_news (FILE *outputfile)
{
	DIR *dir;
	struct dirent *entry;
	uostr_t dirname=UOSTR_INIT;
	size_t pos;

	uostr_xadd_cstrmulti(&dirname,opts.spoolbase,"/",action,"/outgoing",NULL);
	uostr_x0(&dirname);
	tzset();
	dir = opendir (dirname.data);
	if (!dir) {
		diropen_error(dirname.data);
		return;
	}
	uostr_cut(&dirname,-1); /* \0 weg */
	pos=dirname.len;

	entry = readdir (dir);
	while (entry)
	{
		if (entry->d_name[0] == 'm' && entry->d_name[1]=='t')
		{
			uostr_cut(&dirname,pos);
			uostr_xadd_cstrmulti(&dirname,"/",entry->d_name,NULL);
			uostr_x0(&dirname);
			if (!append_file(dirname.data,outputfile)) {
				log(LOG_NOTICE,_("trouble appending mausbatch %s: %s"), dirname.data,
					strerror(errno));
			} else if (opt_unlink_appended_news)
				unlink(dirname.data);
		}
		entry = readdir (dir);
	}
	closedir (dir);
	uostr_freedata(&dirname);
}


static BOOLEAN
convert_one_file (mio_t *ochan, char *fname, uostr_t *mbuf, metamsg_t *meta, BOOLEAN is_mail)
{
	char *transferidp;

	/*
	 * Die Nachricht aus der Datei einlesen in den Puffer.
	 */
	mbuf->len=0;
	if (0==strcmp(fname,"-")) {
		if (rfc_read_message_fd(0,mbuf)!=0)
			return FALSE;
	} else {
		if (rfc_read_message(fname,mbuf)!=0) {
#ifdef HAVE_LSTAT
				/* Pech gehabt, Nachricht nicht lesbar */
				struct stat sb;
				if (lstat(fname,&sb)==0)
				{
					/* wenn symbolischer Link, dann weg damit */
					if ((sb.st_mode & S_IFMT) == S_IFLNK)
						unlink(fname);
				}
#endif
			return FALSE;
		}
	}
	transferidp=strrchr(fname,'/');
	if (transferidp)
		transferidp++;

	if (rfc_to_meta(mbuf,meta,is_mail))
		/* Fehler beim Umwandeln */
		return FALSE;
	meta->is_mail=is_mail; /* XXX needed? */
	if (transferidp)
		meta->transfer_id=XSTRDUP(transferidp);
	else {
		char num1[STR_ULONG];
		char num2[STR_ULONG];
		str_ulong(num1,time(NULL));
		str_ulong(num2,getpid());
		meta->transfer_id=xconcat(num1,".",num2,NULL);
	}

	return rfc2maus (ochan, meta,fname);
}

static void
convert_mail (mio_t *ochan)
{
	uostr_t dirname=UOSTR_INIT;
	struct dirent *entry;
	DIR *dir;
	uostr_t mbuf=UOSTR_INIT;
	metamsg_t *meta;
	size_t pos;
	meta=meta_alloc();
	XNULLCHECK(meta);

	uostr_xadd_cstrmulti(&dirname,opts.spoolbase,"/",action,"/outgoing",NULL);
	uostr_x0(&dirname);
	tzset();

	dir = opendir (dirname.data);
	if (!dir) {
		diropen_error(dirname.data);
		uostr_freedata(&dirname);
		return;
	}
	uostr_cut(&dirname,-1); /* \0 weg */
	pos=dirname.len;

	while ((entry = readdir (dir))!=NULL)
	{
		if (strncmp(entry->d_name,"mail",4)!=0)
			continue;
		uostr_xadd_cstrmulti(&dirname,"/",entry->d_name,NULL);
		uostr_x0(&dirname);
		convert_one_file(ochan,dirname.data,&mbuf,meta, TRUE);
		uostr_cut(&dirname,pos);
	}
#ifndef CLOSEDIR_VOID
	if (closedir(dir))
	{
		uostr_x0(&dirname);
		fprintf(stderr,_("closedir(%s) failed\n"),dirname.data);
	}
#else
	closedir (dir);
#endif
	uostr_freedata(&mbuf);
	meta_free(meta);
	uostr_freedata(&dirname);
}

static void
convert_news (void)
{
	DIR *dir;
	struct dirent *entry;
	uostr_t dirname=UOSTR_INIT;
	uostr_t newsb=UOSTR_INIT;
	uostr_t mausb=UOSTR_INIT;
	uostr_t probf=UOSTR_INIT;
	uostr_t transferid=UOSTR_INIT;
	uostr_t errbuf=UOSTR_INIT;
	metamsg_t *meta;
	meta=meta_alloc();
	XNULLCHECK(meta);

	uostr_xadd_cstr(&dirname,opts.spoolbase);
	uostr_xadd_char(&dirname,'/');
	uostr_xadd_cstr(&dirname,action);
	uostr_xadd_cstr(&dirname,"/outgoing");
	uostr_x0(&dirname);
	tzset();
	dir = opendir (dirname.data);
	if (!dir) {
		diropen_error(dirname.data);
		return;
	}

	while ((entry = readdir (dir))!=NULL)
	{
		char *transferidp;
		BOOLEAN is_suck=FALSE;
		mio_t *ichan;
		mio_t *ochan=NULL;
		mio_retcode_t ret;
		long problems=0;
		BOOLEAN one_ok=FALSE;
		BOOLEAN may_unlink=FALSE;
		/* Nur Files beginnend mit einer Ziffer oder "nb" interessieren
		 * uns hier.
		 */
		if (! ((entry->d_name[0] == 'n' && entry->d_name[1]=='b')
			|| (entry->d_name[0] == 's' && entry->d_name[1]=='k')
			|| isdigit(entry->d_name[0])))
			continue;
		/* alles, was keine Ziffer am Ende hat, ist falsch */
		if (!(isdigit(entry->d_name[strlen(entry->d_name)-1])))
			continue;
		if (entry->d_name[0] == 's' && entry->d_name[1]=='k')
			is_suck=TRUE;

		/* Pfade erzeugen */
		uostr_cut(&newsb,0);
		uostr_xadd_cstr(&newsb,opts.spoolbase);
		uostr_xadd_char(&newsb,'/');
		uostr_xadd_cstr(&newsb,action);
		uostr_xadd_cstr(&newsb,"/outgoing/");
		uostr_xadd_cstr(&newsb,entry->d_name);
		uostr_x0(&newsb);
		uostr_xdup_uostr(&mausb,&newsb);
		uostr_cut(&transferid,0);
		transferidp=strrchr(mausb.data,'/')+1;
		if (isdigit(*transferidp)) {
			uostr_xadd_cstr(&transferid,"mt");
			uostr_xadd_cstr(&transferid,transferidp);
			uostr_x0(&transferid);
			uostr_cut(&mausb,transferidp-mausb.data);
			uostr_xadd_uostr(&mausb,&transferid);
			may_unlink=TRUE;
		} else {
			transferidp[0]='m';
			transferidp[1]='t';
			uostr_xadd_cstr(&transferid,transferidp);
			uostr_x0(&transferid);
			/* "nbxxx" -> "mtxxx", zwecks Zuordnung bei maus2rfc,
			 * bzw "suxxx" -> "mtxxx", suck-format */
			transferid.data[0]='m';
			transferid.data[1]='t';
		}

		if (is_suck && MIO_OK!=mio_suck.open(newsb.data,"r",&ichan,&errbuf))
		{
			struct stat st;
			/* dangling symlink? */
			if (-1==stat(newsb.data,&st) && errno==ENOENT) {
				unlink(newsb.data);
				continue;
			}
			log(LOG_NOTICE, _("Trouble opening suckbatch %s: %s"), newsb.data,
				errbuf.data);
			continue;
		}
		else if (MIO_OK!=mio_rnews.open(newsb.data,"r",&ichan,&errbuf))
		{
			struct stat st;
			/* dangling symlink? */
			if (-1==stat(newsb.data,&st) && errno==ENOENT) {
				unlink(newsb.data);
				continue;
			}
			log(LOG_NOTICE,_("Trouble opening rnewsbatch %s: %s"), newsb.data,
				errbuf.data);
			continue;
		}


		do 
		{
			ret=ichan->fn->import(ichan,meta);
			if (ret==MIO_EOF)
				break;
			if (ret==MIO_ERROR)
			{
				log(LOG_NOTICE, _("%sbatch %s corrupt: %s"), 
					is_suck ? "suck" : "rnews",
					newsb.data,
					ichan->errtext.data ? ichan->errtext.data : _("unknown error"));
				problems++;
				break;
			}
			else if (ret==MIO_RETRY)
			{
				log(LOG_NOTICE,_("%sbatch %s: %s"), 
					is_suck ? "suck" : "rnews",
					newsb.data,
					ichan->errtext.data ? ichan->errtext.data : _("unknown error"));
				if (meta->message_id)
					msglog("%s: %s\n",meta->message_id,
						ichan->errtext.data ? ichan->errtext.data : _("unknown error"));
				problems++;
			}
			else /* MIO_OK */
			{
				if (!ochan && MIO_OK!=mio_maustausch.open(mausb.data,"w",&ochan,&errbuf))
				{
					log(LOG_NOTICE,_("Trouble opening mausbatch %s: %s"), 
						newsb.data, errbuf.data);
					exit(1);  /* Weitermachen sinnlos. */
				}
				meta->transfer_id=XSTRDUP(transferid.data);
				one_ok|=rfc2maus (ochan, meta, may_unlink ? newsb.data : NULL);
			}
		} while (TRUE);

		if (problems!=0)
		{
			uostr_dup_cstrmulti(&probf,opts.spoolbase,"/",action,"/problems/",entry->d_name,NULL);
			uostr_x0(&probf);
			rename(newsb.data,probf.data);


		}
		if (ochan && !one_ok)
			unlink(mausb.data);
		/* im Usermodus die Nachricht nicht lschen - wir brauchen
		 * sie noch um sie canceln zu knnen wenn opts.cancel_original
		 * TRUE ist.
		 * Das ist allerdings groer Bldsinn wenn CNFS im Spiel ist.
		 * Dann bekommt man keine einzelnen Nachrichten mehr sondern
		 * nur noch rnews-Batches.
		 */
		if (ret!=MIO_ERROR && opts.workmode!=USER && opts.cancel_original)
			unlink(newsb.data);
		else if (opt_unlink_news && ret!=MIO_ERROR)
			unlink(newsb.data);
		ichan->fn->close(ichan);
		if (ochan)
			ochan->fn->close(ochan);
	}
	closedir (dir);
	uostr_freedata(&newsb);
	uostr_freedata(&mausb);
	uostr_freedata(&probf);
	uostr_freedata(&dirname);
	uostr_freedata(&transferid);
	uostr_freedata(&errbuf);
	meta_free(meta);
}

static void
quark_mail2spool(const char *user,time_t since)
{
	uostr_t mbuf=UOSTR_INIT;
	metamsg_t *meta;
	uostr_t fname=UOSTR_INIT;
	meta=meta_alloc();
	XNULLCHECK(meta);

	if (user==NULL || !*user) {
		DIR *dir;
		struct dirent *e;
		size_t pos;
		dir=opendir(opts.quark_user_dir);
		if (!dir) {
			log(LOG_NOTICE,"cannot open %s: %s",opts.quark_user_dir,strerror(errno));
			exit(1);
		}
		uostr_xadd_cstr(&fname,opts.quark_user_dir);
		pos=fname.len;
		while ((e=readdir(dir))!=NULL) {
			struct stat st;
			if (*e->d_name=='.')
				continue;
			uostr_xadd_cstr(&fname,e->d_name);
			uostr_x0(&fname);
			if (stat(fname.data,&st)==0 && S_ISDIR(st.st_mode)) {
				quark_mail2spool_oneuser(fname.data,since,&mbuf,meta);
			}
			uostr_cut(&fname,pos);
		}
		closedir(dir);
	} else {
		char *p;
		uostr_xadd_cstr(&fname,opts.quark_user_dir);
		uostr_xadd_cstr(&fname,user);
		uostr_x0(&fname);
		p=fname.data;
		str_lwr(p);
		while (*p) {
			if (isspace(*p) || *p=='/')
				*p='_';
			p++;
		}
		quark_mail2spool_oneuser(fname.data,since,&mbuf,meta);
	}
	meta_free(meta);
	uostr_freedata(&fname);
}

	/* Verfahren:
	 *   if (since==-1)
	 *     convert Maildir/new/{*} und move nach cur.
	 *   else
	 *      if (datum im Dateinamen >since)
	 *         convert it
	 * Ergo: Mailreader sollten da nichts dran ndern ...
	 */
static void
quark_mail2spool_oneuser(const char *dir,time_t since,uostr_t *mbuf,metamsg_t *meta)
{
	struct stat st;
	uostr_t dirname=UOSTR_INIT;
	FILE *new_db;
	uostr_t dbfname1=UOSTR_INIT;
	uostr_t dbfname2=UOSTR_INIT;
	uostr_t outputfname=UOSTR_INIT;
	uostr_t errbuf=UOSTR_INIT;
	void *hash_handle;
	char numbuf[STR_ULONG];
	int i;
	mio_t *ochan;
	
	/* init database of already converted files */
	uostr_xdup_cstrmulti(&dirname,dir,"/Maildir/",NULL);
	uostr_x0(&dirname);
	new_db=quark_mail2spool_loaddb(dirname.data, &dbfname1, &dbfname2, &hash_handle);
	if (!new_db)
		goto cleanup;

	/* Outputfile ffnen 
	 * zuerst wird die Datei ins tmp gelegt, spter 
	 * ins outgoing gelinkt. Das ist frei von unntigen
	 * raceconditions.
	 */
	/* erst mal das Verzeichnis anlegen */
	uostr_xadd_cstrmulti(&outputfname,dir,"/tmp",NULL);
	uostr_x0(&outputfname);
	if (0!=stat(outputfname.data,&st) && errno==ENOENT)
		mkdir(outputfname.data,0700);
	uostr_cut(&outputfname,-1);
	uostr_xadd_cstr(&outputfname,"/mtmail.");
	str_ulong(numbuf,(unsigned long)getpid());
	uostr_xadd_cstr(&outputfname,numbuf);
	uostr_x0(&outputfname);

	if (MIO_OK!=mio_maustausch.open(outputfname.data, "w", &ochan, &errbuf))
	{
		fprintf (stderr, "rfc2maus: %s: %s\n", outputfname.data,errbuf.data);
		log (LOG_ERR, "rfc2maus: %s: %s", outputfname.data,errbuf.data);
		goto cleanup;
	}

	for (i=0;i<2;i++) { /* pass 1: new. pass 2: cur */
		DIR *d;
		struct dirent *e;

		uostr_xdup_cstrmulti(&dirname,dir,"/Maildir/",NULL);
		if (i==0)
			uostr_xadd_cstr(&dirname,"new/");
		else
			uostr_xadd_cstr(&dirname,"cur/");
		uostr_x0(&dirname);

		d=opendir(dirname.data);
		if (!d) {
			log(LOG_NOTICE,"cannot open %s: %s",dirname.data,strerror(errno));
			continue;
		}
		uostr_cut(&dirname,-1); /* \0 wieder weg */

		uostr_xadd_char(&dirname,'/');
		while ((e=readdir(d))!=NULL) {
			size_t pos;
			DYN_ENTRY item,*found;
			if (e->d_name[0]=='.')
				continue;
			item.u.nr=0;
			item.crc=crc32ansi(e->d_name,strlen(e->d_name));
			found=hhsearch(hash_handle,&item,DYN_FIND);
			if (found) {
				if (-1!=since)
					continue; /* already done it */
				if (0!=since) {
					unsigned long t=strtol(e->d_name,NULL,10);
					if (0==t)
						continue;
					if (t<(unsigned long)since)
						continue;
				}
			}
			pos=dirname.len;
			uostr_xadd_cstr(&dirname,e->d_name);
			uostr_x0(&dirname);
			if (stat(dirname.data,&st)==0 && S_ISREG(st.st_mode)) {
				if (convert_one_file(ochan,dirname.data,mbuf,meta,TRUE))
				{
					if (!found) {
						fprintf(new_db,"%s\n",e->d_name);
						found=hhsearch(hash_handle,&item,DYN_OVERWRITE);
						XNULLCHECK(found);
					}
				}
			} 
			uostr_cut(&dirname,pos);
		}
		/* done this dir */
#ifndef CLOSEDIR_VOID
		if (closedir(d))
		{
			uostr_x0(&dirname);
			fprintf(stderr,_("closedir(%s) failed\n"),dirname.data);
			uostr_cut(&dirname,-1);
		}
#else
		closedir (d);
#endif
	} /* for loop */
	/* cur and new dir done */

	ochan->fn->close(ochan);	
	if (stat(outputfname.data,&st)==0 && st.st_size==0)
		unlink(outputfname.data);
	else {
		/* move outputfile from tmp to outgoing */
		uostr_t renname=UOSTR_INIT;
		int num;
		uostr_xdup_cstrmulti(&renname,dir,"/outgoing",NULL);
		uostr_x0(&renname);
		if (0!=stat(renname.data,&st) && errno==ENOENT)
			mkdir(renname.data,0700);
		uostr_cut(&renname,-1);
		uostr_xadd_cstr(&renname,"/mtmail.");
		for (num=0;;num++) {
			int r;
			size_t pos;
			pos=renname.len;
			str_ulong(numbuf,(unsigned long)num);
			uostr_xadd_cstr(&renname,numbuf);
			uostr_x0(&renname);
			r=stat(renname.data,&st);
			if (r!=0) {
				if (errno==ENOENT)
					break;
				else {
					log(LOG_NOTICE,"cannot stat %s: %s",renname.data,strerror(errno));
				}
			}
			uostr_cut(&renname,pos);
		}
		if (-1==link(outputfname.data,renname.data)) {
			log(LOG_NOTICE,"cannot link %s to %s: %s",outputfname.data,renname.data,strerror(errno));
		}
		unlink(outputfname.data);
		uostr_freedata(&renname);
	}

	quark_mail2spool_savedb(&dbfname1,&dbfname2,hash_handle,new_db);
  cleanup:
	uostr_freedata(&errbuf);
	uostr_freedata(&dbfname1);
	uostr_freedata(&dbfname2);
	uostr_freedata(&dirname);
	uostr_freedata(&outputfname);
}

static void
quark_mail2spool_savedb(uostr_t *dbfname1, uostr_t *dbfname2,void *handle, FILE *f)
{
	struct stat st;
	hhdestroy(handle);
	if (f)
		fclose(f);
	if (0!=stat(dbfname2->data,&st) || st.st_size==0) {
		unlink(dbfname2->data);
	} else  {
		unlink(dbfname1->data);
		if (-1==rename(dbfname2->data,dbfname1->data)) {
			/* Ugh */
			log(LOG_NOTICE,"cannot rename %s to %s: %s",dbfname1->data,dbfname2->data,strerror(errno));
		}
	}
}

static FILE * 
quark_mail2spool_loaddb(const char *dir, uostr_t *dbfname1, uostr_t *dbfname2, void **handle)
{
	FILE *f,*newf;
	char buf[256]; /* filename in a maildir is never longer */
	uostr_t md=UOSTR_INIT; /* see below. is the file still there? */
	size_t md_pos;

	*handle=hhcreate(9511);
	XNULLCHECK(*handle);

	uostr_xdup_cstrmulti(dbfname1,dir,"/.qm2ssentdb",NULL);
	uostr_x0(dbfname1);
	f=fopen(dbfname1->data,"r");

	uostr_xdup_uostr(dbfname2,dbfname1);
	uostr_cut(dbfname2,-1);
	uostr_xadd_cstr(dbfname2,".new");
	uostr_x0(dbfname2);
	newf=fopen(dbfname2->data,"w");
	if (!newf) {
		struct stat st;
		int e=errno;
		if (f) {
			fclose(f);
		}
		if (-1!=stat(dir,&st) || errno!=ENOENT)
			log(LOG_NOTICE,"cannot open converter database in %s",dir);
		errno=e;
		hhdestroy(*handle);
		*handle=NULL;
		return NULL;
	}
	if (!f) 
		return newf;


	uostr_xdup_cstr(&md,dir);
	md_pos=md.len;
	while (fgets(buf,sizeof(buf),f)) {
		DYN_ENTRY item, *found;
		char c;
		struct stat st;
		char *p=strpbrk(buf,"\r\n");
		if (!p)
			break;
		c=*p;
		*p=0;

		uostr_cut(&md,md_pos);
		uostr_xadd_cstrmulti(&md,"/new/",buf,NULL);
		uostr_x0(&md);
		if (0!=stat(md.data,&st)) {
			uostr_cut(&md,md_pos);
			uostr_xadd_cstrmulti(&md,"/cur/",buf,NULL);
			uostr_x0(&md);
			if (0!=stat(md.data,&st))
				/* forget about this entry, the file has been deleted */
				continue;
		}
		/* file still exists */
		item.crc=crc32ansi(buf,p-buf);
		item.u.nr=0;
		found=hhsearch(*handle,&item,DYN_OVERWRITE);
		XNULLCHECK(found);
		*p=c;
		fputs(buf,newf);
	}
	fclose(f);
	uostr_freedata(&md);
	return newf;
}
